import numpy as np
import jax.numpy as jnp
import haiku as hk

from .base import QuantumSystem
from .circuit import CircuitLCJ
from ..utils.utility import const_init
from ..utils.operators import destroy, create, qeye


class Fluxonium(CircuitLCJ):
    """Class for the fluxonium qubit.

    When considering time-varying external flux,
    the phiext should be on EL. See PRB 99, 174512 (2019)
    Note that Transmon EL=0.

    Args:
        ec (float, optional): charging energy in unit GHz
        ej (float, optional): Josephson energy in unit GHz
        el (float, optional): inductive energy in unit GHz
        constant (bool): True for manually setting parameters, False for using
            haiku's parameters management.
        phiext: phi external parameters for fluxonium, in unit of flux quanta
        put_phiext_on_inductor: Whether to put the external flux term
            in the EL term (True) or EJ term (False).
        num_basis: Number of basis functions
        truncated_dim: Number of eigenenergies to construct the Hilbert space.
        basis: Basis set, can be {'charge', 'phase', 'phase_only'}.
        n_max: charge basis range [-n_max, n_max).
                Override phi_max if both specified
        phi_max: phi basis range [-phi_max, phi_max), must be multiple of pi
        is_basis_sym: Is basis forced to be [-n, ... +n]
            instead of [-n, ... +n-1].
        name: module name
        var: the device parameters variance
        drive_for_state_phase: This is to specify which drive is used, so
            eigenstates are rotated to let <i|drive|i+1> be real positive
            Possible values are "charge" and "phase". Default is "charge".
    """

    def __init__(self,
                 ec: float = None,
                 ej: float = None,
                 el: float = None,
                 constant: bool = False,
                 phiext: float = None,
                 put_phiext_on_inductor: bool = True,
                 num_basis: int = 400,
                 truncated_dim: int = 10,
                 basis: str = 'phase',
                 n_max: int = 0,
                 phi_max: float = 5 * np.pi,
                 is_basis_sym: bool = False,
                 name: str = 'fluxonium',
                 var: dict = None,
                 drive_for_state_phase: str = 'charge',
                 **kwargs) -> None:
        super().__init__(False, basis, num_basis, truncated_dim, n_max, phi_max,
                         is_basis_sym, drive_for_state_phase, name)

        if not constant:
            self._ec = hk.get_parameter('ec', [], init=const_init(ec))
            self._ej = hk.get_parameter('ej', [], init=const_init(ej))
            self._el = hk.get_parameter('el', [], init=const_init(el))
            # add gaussian distribution variance
            self.add_lcj_params_variance(var)
        else:
            self.ec = ec
            self.ej = ej
            self.el = el
        if phiext is not None:
            self.phiext = phiext
        else:
            self.phiext = hk.get_parameter('phiext', [], init=jnp.zeros)
        if basis not in ['phase', 'charge', 'phase_only']:
            raise NotImplementedError(f'Do not support basis {basis}.')

        self.put_phiext_on_inductor = put_phiext_on_inductor

    def add_lcj_params_variance(self, var=None):
        """Function to add variance to (device parameters).
        """
        if isinstance(var, dict):
            self.ec = self._ec * var.get('ec', 1.0)
            self.ej = self._ej * var.get('ej', 1.0)
            self.el = self._el * var.get('el', 1.0)
        else:
            self.ec = self._ec
            self.ej = self._ej
            self.el = self._el

    def create_t(self) -> jnp.ndarray:
        """Computes kinetic matrix in phase/charge basis.

        Returns:
            the kinetic matrix
        """
        # Ec
        if self.basis == "charge":
            return jnp.diag(self.ec * 4 * self.ar_n**2)
        # Use central difference
        elif self.basis == "phase_only":
            m = -self.create_d2phi()
            return m * (self.ec * 4)
        elif self.basis == "phase":
            n_square = self.transform_n_to_phi(jnp.diag(self.ar_n**2))
            return n_square * (self.ec * 4)

    def create_v(self) -> jnp.ndarray:
        """Computes potential matrix.

        Returns:
            potential matrix.
        """
        # Create the matrix in phase basis
        if self.put_phiext_on_inductor:
            # EJ
            m = -1.0 * self.ej * jnp.cos(self.ar_phi)
            # EL
            m += self.el * (self.ar_phi + self.phiext)**2 / 2
        else:
            # EJ
            m = -1.0 * self.ej * jnp.cos(self.ar_phi - self.phiext)
            # EL
            m += self.el * (self.ar_phi)**2 / 2

        if self.basis == "charge":
            m = self.transform_phi_to_n(jnp.diag(m))
            return m
        elif self.basis in ("phase", "phase_only"):
            return jnp.diag(m)

    def idling_hamiltonian(self) -> jnp.ndarray:
        """Create the Hamiltonian matrix of the qubit.

        Returns:
            ndarray of the Hamiltonian matrix
        """

        if self.basis in ("phase", "charge", "phase_only"):
            mt = self.create_t()
            mv = self.create_v()
            return mt + mv
        else:
            raise ValueError(f"Unknown basis {self.basis}")

    def n_operator(self, **kwargs) -> jnp.ndarray:
        """
        Return:
            Returns the :math:`n = - i d/d\\phi` operator in the phase basis.
        """

        if self.eigvec_in_raw_basis is None:
            self.eigenenergies()
        n_in_raw_basis = self.create_n()
        n_eigenbasis = jnp.conj(self.eigvec_in_raw_basis
                               ).T @ n_in_raw_basis @ self.eigvec_in_raw_basis

        return n_eigenbasis[:self.dim, :self.dim]

    def phi_operator(self, **kwargs) -> jnp.ndarray:
        """
        Return:
            Returns the phi operator in the phase basis.
        """

        if self.eigvec_in_raw_basis is None:
            self.eigenenergies()
        phi_in_raw_basis = self.create_phi()
        phi_eigenbasis = jnp.conj(
            self.eigvec_in_raw_basis
        ).T @ phi_in_raw_basis @ self.eigvec_in_raw_basis
        phi_truncated_eigenbasis = phi_eigenbasis[:self.dim, :self.dim]
        if self.put_phiext_on_inductor:
            phi_truncated_eigenbasis += jnp.eye(
                phi_truncated_eigenbasis.shape[0]) * self.phiext

        return phi_truncated_eigenbasis


class Transmon(CircuitLCJ):
    """Class for the transmon qubit.

    Args:
        ec (float, optional): charging energy
        ej (float, optional): Josephson energy
        ng (float, optional): ng parameter for transmon
            set `None` if you want to pass parameters by dm-haiku.
        constant (bool): True for manually setting parameters, False for using
            haiku's parameters management.
        d (float, optional): junction asymmetry parameter,
            set `None` if it's not a Tunable Transmon.
        phiext: phi external parameters for fluxonium, in unit of flux quanta
        num_basis: Number of basis functions
        truncated_dim: Number of eigenenergies to construct the Hilbert space.
        basis: Basis set, can be {'charge', 'phase', 'phase_only'}.
        n_max: charge basis range [-n_max, n_max).
                Override phi_max if both specified
        is_basis_sym: Is basis forced to be [-n, ... +n]
            instead of [-n, ... +n-1].
        name: module name
        var: the device parameters variance
        drive_for_state_phase: This is to specify which drive is used, so
            eigenstates are rotated to let <i|drive|i+1> be real positive
            Possible values are "charge" and "phase". Default is "charge".
    """

    def __init__(self,
                 ec: float = None,
                 ej: float = None,
                 ng: float = None,
                 constant: bool = False,
                 d: float = None,
                 phiext: float = None,
                 num_basis: int = 400,
                 truncated_dim: int = 10,
                 basis: str = 'phase',
                 n_max: int = 0,
                 is_basis_sym: bool = False,
                 name: str = 'transmon',
                 var: jnp.ndarray = None,
                 drive_for_state_phase: str = 'charge',
                 **kwargs) -> None:
        phi_max = np.pi  # Periodicity of 2pi
        super().__init__(True, basis, num_basis, truncated_dim, n_max, phi_max,
                         is_basis_sym, drive_for_state_phase, name)

        if not constant:
            self._ec = hk.get_parameter('ec', [], init=const_init(ec))
            self._ej = hk.get_parameter('ej', [], init=const_init(ej))
            self._el = 0
            self.add_lcj_params_variance(var)
        else:
            self.ec = ec
            self.ej = ej
        if ng is None:
            self.ng = hk.get_parameter('ng', [], init=jnp.ones)
        else:
            self.ng = ng
        if d is None:
            self.tunable = False
        else:
            self.tunable = True
            self.d = d
        if basis not in ['phase', 'charge', 'phase_only']:
            raise NotImplementedError(f'Do not support basis {basis}.')

        if phiext is not None:
            self.phiext = phiext
        else:
            self.phiext = hk.get_parameter('phiext', [], init=jnp.zeros)
        self.put_phiext_on_inductor = False  # Transmon do not have inductor

    def add_lcj_params_variance(self, var=None):
        """Function to add variance to (device parameters).
        """
        if isinstance(var, dict):
            self.ec = self._ec * var.get('ec', 1.0)
            self.ej = self._ej * var.get('ej', 1.0)
        else:
            self.ec = self._ec
            self.ej = self._ej

    def create_t(self) -> jnp.ndarray:
        """Computes kinetic matrix in phase/charge basis.

        Returns:
            the kinetic matrix
        """
        # Ec
        if self.basis == "charge":
            return jnp.diag(self.ec * 4 * (self.ar_n - self.ng)**2)
        # Use central difference
        elif self.basis == "phase_only":
            m = -self.create_d2phi()
            m += 2j * self.ng * self.create_dphi() + np.eye(
                self.ar_phi.size) * self.ng**2
            return m * (self.ec * 4)
        elif self.basis == "phase":
            n1 = self.transform_n_to_phi(jnp.diag(self.ar_n))
            n_square = self.transform_n_to_phi(jnp.diag(self.ar_n**2))
            m = n_square - 2 * self.ng * n1 + jnp.eye(
                self.ar_phi.size) * self.ng**2
            return m * (self.ec * 4)

    def create_v(self) -> jnp.ndarray:
        """Computes potential matrix.

        Returns:
            potential matrix.
        """
        if self.tunable:
            ej_eff = self.ej * jnp.sqrt(
                jnp.cos(self.phiext / 2.0)**2 +
                self.d**2 * jnp.sin(self.phiext / 2.0)**2)
        else:
            ej_eff = self.ej

        if self.basis == "charge":
            m = -1.0 * ej_eff * self.charge_cos
            return m
        elif self.basis in ("phase", "phase_only"):
            m = -1.0 * ej_eff * jnp.cos(self.ar_phi)  # EJ
            return jnp.diag(m)

    def idling_hamiltonian(self) -> jnp.ndarray:
        """Create the Hamiltonian matrix of the qubit.

        Returns:
            ndarray of the Hamiltonian matrix
        """

        if self.basis in ("phase", "charge", "phase_only"):
            mt = self.create_t()
            mv = self.create_v()
            return mt + mv
        else:
            raise ValueError(f"Unknown basis {self.basis}")

    def n_operator(self, **kwargs) -> jnp.ndarray:
        """
        Return:
            Returns the :math:`n = - i d/d\\phi` operator in the phase basis.
        """

        if self.eigvec_in_raw_basis is None:
            self.eigenenergies()
        n_in_raw_basis = self.create_n()
        n_eigenbasis = jnp.conj(self.eigvec_in_raw_basis
                               ).T @ n_in_raw_basis @ self.eigvec_in_raw_basis

        return n_eigenbasis[:self.dim, :self.dim]

    def phi_operator(self, **kwargs) -> jnp.ndarray:
        """
        Return:
            Returns the phi operator in the phase basis.
        """

        if self.eigvec_in_raw_basis is None:
            self.eigenenergies()
        phi_in_raw_basis = self.create_phi()
        phi_eigenbasis = jnp.conj(
            self.eigvec_in_raw_basis
        ).T @ phi_in_raw_basis @ self.eigvec_in_raw_basis
        phi_truncated_eigenbasis = phi_eigenbasis[:self.dim, :self.dim]

        return phi_truncated_eigenbasis


class Resonator(QuantumSystem):
    """Class representing a harmonic resonator.

    Args:
        f_res: resonator frequency in unit GHz
        remove_zpe: Remove the Zero-point energy(ZPE) of the resonator.
        constant (bool): True for manually setting parameters, False for using
            haiku's parameters management.
        truncated_dim: Number of eigenenergies to construct the Hilbert space.
        name: module name
    """

    def __init__(self,
                 f_res: float = None,
                 remove_zpe: bool = False,
                 constant: bool = False,
                 truncated_dim: int = 10,
                 name: str = 'resonator',
                 **kwargs) -> None:
        super().__init__(name=name)

        if not constant:
            self.f_res = hk.get_parameter('f_res', [], init=const_init(f_res))
        else:
            self.f_res = f_res
        self.remove_zpe = remove_zpe
        # Truncated parameter
        self.truncated_dim = truncated_dim
        # Cached variable
        self.eigval = None
        self.eigvec_in_raw_basis = None

    @property
    def qdevice_type(self):
        return 'resonator'

    def idling_hamiltonian(self):
        """Create the hamiltonian matrix of the resonator.

        Returns:
            ndarray of the Hamiltonian matrix
        """
        a = destroy(self.truncated_dim)
        adag = create(self.truncated_dim)
        if self.remove_zpe:
            ham = adag @ a
        else:
            ham = adag @ a + 0.5 * qeye(self.truncated_dim)
        return ham * self.f_res

    def eigenenergies(self):
        """Returns array of eigenvalues.

        Args:
            evals_count:
                number of desired eigenvalues (default value = 6)
        """
        evals = self._calc_evals()
        return evals[:self.dim]

    @property
    def dim(self) -> int:
        """Returns truncated Hilbert space dimension"""
        return self.truncated_dim

    def creation_operator(self, **kwargs) -> jnp.ndarray:
        """Returns the creation operator"""
        return create(self.dim)

    def annihilation_operator(self, **kwargs) -> jnp.ndarray:
        """Returns the creation operator"""
        return destroy(self.dim)



import numpy as np
import scipy.linalg as sp

from supergrad.quantum_system import KronObj, LindbladObj

from jax import config

config.update('jax_enable_x64', True)

rng = np.random.default_rng(seed=42)
dims = [3, 2, 4]
mat0 = rng.random((dims[0], dims[0]))
mat0_1 = rng.random((dims[0], dims[0]))
mat1 = rng.random((dims[1], dims[1]))
mat2 = rng.random((dims[2], dims[2]))
kron1 = KronObj([mat0, mat1], dims, [0, 1])
manual1 = np.kron(mat0, mat1)
manual1 = np.kron(manual1, np.eye(dims[2]))

mat22 = rng.random((dims[0], dims[0]))
mat3 = rng.random((dims[2], dims[2]))
kron2 = KronObj([mat22, mat3], dims, [0, 2])
manual2 = np.kron(mat22, np.eye(dims[1]))
manual2 = np.kron(manual2, mat3)

mat4 = rng.random((dims[1], dims[1]))
kron3 = KronObj([mat4], dims, [1])
manual3 = np.kron(np.eye(dims[0]), mat4)
manual3 = np.kron(manual3, np.eye(dims[2]))

ident = np.eye(manual1.shape[0], dtype=complex)

density = rng.random((np.prod(dims), np.prod(dims)))

lind1 = LindbladObj()
lind1.add_liouvillian(kron1)
lind2 = LindbladObj()
lind2.add_lindblad_operator(kron2)
lind3 = LindbladObj()
chi3 = 0.3
lind3.add_lindblad_operator(kron2, kron1 + kron3, chi3)

# Convert Hamiltonian to Liouvillian
manual_lind1 = -1.0j * (np.kron(ident, manual1) -
                        np.kron(manual1.transpose(), ident))
# Convert collapse opeartor to dissipator
ad_b = manual2.conjugate().transpose() @ manual2
manual_lind2 = np.kron(manual2.conjugate(), manual2) - 0.5 * np.kron(
    ident, ad_b) - 0.5 * np.kron(ad_b.transpose(), ident)

ad_b = manual2.conjugate().transpose() @ (manual1 + manual3)
manual_lind3 = np.kron((manual1 + manual3).conjugate(), manual2) * np.exp(
    1j * chi3) - 0.5 * np.kron(ident, ad_b) - 0.5 * np.kron(
        ad_b.transpose(), ident)


def test_liouvillian():
    assert np.allclose(manual_lind1, lind1.full().reshape(manual_lind1.shape))


def test_dissipator():
    assert np.allclose(manual_lind2, lind2.full().reshape(manual_lind2.shape))
    assert np.allclose(manual_lind3, lind3.full().reshape(manual_lind3.shape))


def test_tensor_multiplication():
    res = (manual_lind1 @ density.ravel('F')).reshape(np.prod(dims),
                                                      np.prod(dims)).T
    res0 = lind1 @ density
    assert np.allclose(res0, res)
    res = (manual_lind2 @ density.ravel('F')).reshape(np.prod(dims),
                                                      np.prod(dims)).T
    res0 = lind2 @ density
    assert np.allclose(res0, res)
    res = (manual_lind3 @ density.ravel('F')).reshape(np.prod(dims),
                                                      np.prod(dims)).T
    res0 = lind3 @ density
    assert np.allclose(res0, res)


def test_tensor_multiplication_unit1():
    lind_test = LindbladObj([[mat0, mat2], [mat1]], dims, [[0, 2], [1]])
    manual_test = np.kron(mat0, np.eye(dims[1]))
    manual_test = np.kron(manual_test, mat2)
    manual_test_1 = np.kron(np.eye(dims[0]), mat1)
    manual_test_1 = np.kron(manual_test_1, np.eye(dims[2]))
    manual_lind_test = np.kron(manual_test, manual_test_1)
    res = (manual_lind_test @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0 = lind_test @ density
    res_exp = (sp.expm(manual_lind_test) @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0_exp = lind_test.expm(density)
    assert np.allclose(res0_exp, res_exp)
    assert np.allclose(manual_lind_test,
                       lind_test.full().reshape(manual_lind_test.shape))
    assert np.allclose(res0, res)


def test_tensor_multiplication_unit2():
    lind_test = LindbladObj([[mat0], [mat0_1, mat1]], dims, [[0], [0, 1]])
    manual_test = np.kron(mat0, np.eye(dims[1]))
    manual_test = np.kron(manual_test, np.eye(dims[2]))
    manual_test_1 = np.kron(mat0_1, mat1)
    manual_test_1 = np.kron(manual_test_1, np.eye(dims[2]))
    manual_lind_test = np.kron(manual_test, manual_test_1)
    res = (manual_lind_test @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0 = lind_test @ density
    res_exp = (sp.expm(manual_lind_test) @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0_exp = lind_test.expm(density, 1)
    assert np.allclose(res0_exp, res_exp)
    assert np.allclose(manual_lind_test,
                       lind_test.full().reshape(manual_lind_test.shape))
    assert np.allclose(res0, res)


def test_tensor_multiplication_unit3():
    lind_test = LindbladObj([[mat0], [mat1]], dims, [[0], [1]])
    manual_test = np.kron(mat0, np.eye(dims[1]))
    manual_test = np.kron(manual_test, np.eye(dims[2]))
    manual_test_1 = np.kron(np.eye(dims[0]), mat1)
    manual_test_1 = np.kron(manual_test_1, np.eye(dims[2]))
    manual_lind_test = np.kron(manual_test, manual_test_1)
    res = (manual_lind_test @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0 = lind_test @ density
    res_exp = (sp.expm(manual_lind_test) @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0_exp = lind_test.expm(density, 1)
    assert np.allclose(res0_exp, res_exp)
    assert np.allclose(manual_lind_test,
                       lind_test.full().reshape(manual_lind_test.shape))
    assert np.allclose(res0, res)


def test_tensor_multiplication_unit4():
    lind_test = LindbladObj([[mat0_1], [mat0]], dims, [[0], [0]])
    manual_test = np.kron(mat0, np.eye(dims[1]))
    manual_test = np.kron(manual_test, np.eye(dims[2]))
    manual_test_1 = np.kron(mat0_1, np.eye(dims[1]))
    manual_test_1 = np.kron(manual_test_1, np.eye(dims[2]))
    manual_lind_test = np.kron(manual_test_1, manual_test)
    res = (manual_lind_test @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0 = lind_test @ density
    res_exp = (sp.expm(manual_lind_test) @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0_exp = lind_test.expm(density, 1)
    assert np.allclose(res0_exp, res_exp)
    assert np.allclose(manual_lind_test,
                       lind_test.full().reshape(manual_lind_test.shape))
    assert np.allclose(res0, res)


def test_tensor_multiplication_unit5():
    lind_test = LindbladObj([[], [mat0]], dims, [[], [0]])
    manual_test = np.kron(mat0, np.eye(dims[1]))
    manual_test = np.kron(manual_test, np.eye(dims[2]))
    manual_lind_test = np.kron(ident, manual_test)
    res = (manual_lind_test @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0 = lind_test @ density
    res_exp = (sp.expm(manual_lind_test) @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0_exp = lind_test.expm(density, 1)
    assert np.allclose(res0_exp, res_exp)
    assert np.allclose(manual_lind_test,
                       lind_test.full().reshape(manual_lind_test.shape))
    assert np.allclose(res0, res)


def test_tensor_multiplication_unit6():
    lind_test = LindbladObj([[mat0, mat1, mat2], []], dims, [[0, 1, 2], []])
    manual_test = np.kron(mat0, mat1)
    manual_test = np.kron(manual_test, mat2)
    manual_lind_test = np.kron(manual_test, ident)
    res = (manual_lind_test @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0 = lind_test @ density
    res_exp = (sp.expm(manual_lind_test) @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0_exp = lind_test.expm(density, 1)
    assert np.allclose(res0_exp, res_exp)
    assert np.allclose(manual_lind_test,
                       lind_test.full().reshape(manual_lind_test.shape))
    assert np.allclose(res0, res)


def test_tensor_multiplication_unit7():
    lind_test = LindbladObj([[mat0, mat1], [mat0, mat1]], dims,
                            [[0, 1], [0, 1]])
    manual_test = np.kron(mat0, mat1)
    manual_test = np.kron(manual_test, np.eye(dims[2]))
    manual_lind_test = np.kron(manual_test, manual_test)
    res = (manual_lind_test @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0 = lind_test @ density
    res_exp = (sp.expm(manual_lind_test) @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0_exp = lind_test.expm(density, 1)
    assert np.allclose(res0_exp, res_exp)
    assert np.allclose(manual_lind_test,
                       lind_test.full().reshape(manual_lind_test.shape))
    assert np.allclose(res0, res)


def test_tensor_multiplication_unit8():
    lind_test = LindbladObj([[mat0, mat1], []], dims, [[0, 1], []])
    manual_test = np.kron(mat0, mat1)
    manual_test = np.kron(manual_test, np.eye(dims[2]))
    manual_lind_test = np.kron(manual_test, ident)
    res = (manual_lind_test @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0 = lind_test @ density
    res_exp = (sp.expm(manual_lind_test) @ density.ravel('F')).reshape(
        np.prod(dims), np.prod(dims)).T
    res0_exp = lind_test.expm(density, 1)
    assert np.allclose(res0_exp, res_exp)
    assert np.allclose(manual_lind_test,
                       lind_test.full().reshape(manual_lind_test.shape))
    assert np.allclose(res0, res)


def test_addition():
    assert np.allclose(manual_lind1 + manual_lind2,
                       (lind1 + lind2).full().reshape(manual_lind1.shape))
    assert np.allclose(manual_lind1 + manual_lind2,
                       (lind2 + lind1).full().reshape(manual_lind1.shape))
    assert np.allclose(manual_lind2 - manual_lind3,
                       (lind2 - lind3).full().reshape(manual_lind1.shape))
    assert np.allclose(manual_lind2 - manual_lind1,
                       (lind2 - lind1).full().reshape(manual_lind1.shape))


def test_scalar_multiplication():
    flt = rng.random()
    assert np.allclose(flt * (manual_lind1 + manual_lind2),
                       (flt * (lind1 + lind2)).full().reshape(
                           manual_lind1.shape))
    assert np.allclose(flt * (manual_lind1 + manual_lind2),
                       ((lind1 + lind2) * flt).full().reshape(
                           manual_lind1.shape))



from typing import Callable
from functools import partial
import numpy as np
from scipy.optimize import minimize
import jax
import jax.numpy as jnp
from jax._src import linear_util as lu
from jax._src.api import _jacfwd_unravel, _std_basis, _jvp
from jax._src.api_util import argnums_partial
from jax.flatten_util import ravel_pytree
from jax.example_libraries import optimizers
import pprint

from supergrad.utils.docstring import implements
from supergrad.utils.utility import convert_to_json_compatible


def value_and_gradfwd(fun: Callable, argnums=0):
    """Constructed by analogy to value_and_grad
    """

    def jacfun(*args, **kwargs):
        f = lu.wrap_init(fun, kwargs)
        f_partial, dyn_args = argnums_partial(
            f, argnums, args, require_static_args_hashable=False)
        pushfwd: Callable = partial(_jvp, f_partial, dyn_args)
        y, jac = jax.vmap(pushfwd, out_axes=(None, -1))(_std_basis(dyn_args))
        jac_tree = jax.tree_util.tree_map(partial(_jacfwd_unravel, dyn_args), y,
                                          jac)
        return y, jac_tree[0]

    return jacfun


def scipy_optimize_wrapper(fun: Callable,
                           unflatten: Callable,
                           jac: bool = True,
                           jit: bool = True,
                           fwd_ad: bool = False,
                           logging: bool = False):
    """A wrapper transform function (written in JAX) to scipy compatible version.

    Args:
        fun (callable): input function
        unflatten (callable): a callable for unflattening a 1D vector of the
            same length back to a pytree.
        jac (bool): True for using Gradient-based optimize algorithm,
            False for Gradient-free optimize algorithm.
        jit (bool): True for just-in-time compile.
        fwd_ad(bool): True for using forward-mode auto diff.
        logging (bool): True for logging current loss and parameters.
    """
    if jac:
        if fwd_ad:
            val_grad = value_and_gradfwd(fun)
        else:
            val_grad = jax.value_and_grad(fun)
        if jit:
            val_grad = jax.jit(val_grad)

        def scipy_fun(*args, **kwargs):
            """Decorated scipy function

            Return:
                a tuple (f, g) containing the objective function and the gradient.
            """
            # ravel x0 parameters pytree
            args_list = list(args)
            args_list[0] = unflatten(args[0])
            val, grad = val_grad(*args_list, **kwargs)
            # transform DeviceArray to numpy array
            scipy_val = np.asarray(jax.tree_util.tree_leaves(val),
                                   dtype=np.float64)
            # fix ValueError
            grad = jax.tree_util.tree_map(lambda leaf: leaf.tolist(), grad)
            scipy_grad = np.asarray(jax.tree_util.tree_leaves(grad),
                                    dtype=np.float64)
            if logging:
                global step
                pp = pprint.PrettyPrinter(indent=2)
                print(f'step: {step}')
                print('parameters:')
                pp.pprint(convert_to_json_compatible(args_list[0]))
                print('gradient:')
                pp.pprint(grad)
                print(f'loss: {float(scipy_val):f}')
                step += 1
            return scipy_val, scipy_grad

        return scipy_fun
    else:
        if jit:
            fun = jax.jit(fun)

        def scipy_fun(*args, **kwargs):
            """Decorated scipy function

            Return:
                the objective function
            """
            # ravel x0 parameters pytree
            args_list = list(args)
            args_list[0] = unflatten(args[0])
            val = fun(*args_list, **kwargs)
            # transform DeviceArray to numpy array
            scipy_val = np.asarray(jax.tree_util.tree_leaves(val),
                                   dtype=np.float64)
            if logging:
                pp = pprint.PrettyPrinter(indent=2)
                print('parameters:')
                pp.pprint(convert_to_json_compatible(args_list[0]))
                print(f'loss: {float(scipy_val):f}')
            return scipy_val

        return scipy_fun


def scipy_hess_wrapper(fun: Callable, unflatten: Callable, jit: bool = True):
    """A wrapper transform function (written in JAX) to scipy compatible version.

    Args:
        fun (callable): input function
        unflatten (callable): a callable for unflattening a 1D vector of the
            same length back to a pytree.
        jit (bool): True for just-in-time compile.
    """
    # using a composition of `jacfwd` and `jacrev` to compute dense Hessian
    hess = jax.hessian(fun)
    if jit:
        hess = jax.jit(hess)

    def scipy_hess(*args, **kwargs):
        """Decorated scipy function

        Return:
            (n, n) dimension dense Hessian matrix
        """
        # ravel x0 parameters pytree
        args_list = list(args)
        args_list[0] = unflatten(args[0])
        # calculate Hessian matrix
        hess_matrix = hess(*args_list, **kwargs)
        # fix ValueError
        hess_matrix = jax.tree_util.tree_map(lambda leaf: leaf.tolist(),
                                             hess_matrix)
        # flatten Pytree and reshape
        dim = len(
            jax.tree_util.tree_leaves(
                jax.tree_util.tree_map(lambda leaf: leaf.tolist(),
                                       args_list[0])))
        hess_matrix = jnp.reshape(
            jnp.array(jax.tree_util.tree_leaves(hess_matrix)), (dim, dim))
        return hess_matrix

    return scipy_hess


def scipy_callback_wrapper(fun: Callable, unflatten: Callable):
    """A wrapper transform unflattening callback function's current parameter
    vector `xk` back to Pytree.

    Args:
        fun (callable): input function
        unflatten (callable): a callable for unflattening a 1D vector of the
            same length back to a pytree.
    """

    def unflatten_callback(xk, *args):
        return fun(unflatten(xk), *args)

    return unflatten_callback


@implements(
    minimize, """
New methods are implemented in arguments ``jac`` and ``hess``.
If ``jac`` is 'jax', the gradient vector will be calculate by JAX.
If ``hess`` is 'jax', the Hessian matrix will be calculate by JAX.
jit : bool, optional
    True for enable just-in-time compile.
fwd_ad : bool, optional
    True for using forward-mode auto diff.
logging : bool, optional
    Whether to output logging messages. Defaults to :py:obj:`False`.
""")
def scipy_minimize(fun,
                   x0,
                   args=(),
                   method=None,
                   jac='jax',
                   hess=None,
                   hessp=None,
                   bounds=None,
                   constraints=(),
                   tol=None,
                   callback=None,
                   options=None,
                   *,
                   jit: bool = True,
                   fwd_ad: bool = False,
                   logging: bool = False):
    global step
    step = 0
    scipy_x0, unflatten = ravel_pytree(x0)
    if jac == 'jax':
        jac = True
        scipy_fun = scipy_optimize_wrapper(fun,
                                           unflatten,
                                           jac=True,
                                           jit=jit,
                                           fwd_ad=fwd_ad,
                                           logging=logging)
    else:
        scipy_fun = scipy_optimize_wrapper(fun,
                                           unflatten,
                                           jac=False,
                                           jit=jit,
                                           logging=logging)
    if callback is not None:
        scipy_callback = scipy_callback_wrapper(callback, unflatten)
    else:
        scipy_callback = None
    if hess == 'jax':
        hess = scipy_hess_wrapper(fun, unflatten, jit=jit)
    # Minimize with scipy
    res = minimize(scipy_fun,
                   scipy_x0,
                   args=args,
                   method=method,
                   jac=jac,
                   hess=hess,
                   hessp=hessp,
                   callback=scipy_callback,
                   bounds=bounds,
                   constraints=constraints,
                   tol=tol,
                   options=options)
    # pack the output back into a PyTree
    res['x'] = unflatten(res['x'])
    if 'jac' in res.keys():
        res['jac'] = unflatten(res['jac'])

    return res


def adam_opt(fun, x0, args=(), options={}, jit=True, fwd_ad=False):
    """The JAX-backend implementation of the Adam optimization algorithm.

    Args:
        fun: the cost function f(params, all_params, `**kwargs`)
        x0: initial guess will be optimized iteratively.
        args (tuple, optional): Extra arguments passed to the objective
            function and its derivative. (fun, jac and hess functions)
        options (dict, optional): optimizer's hyper parameters dictionary.
            The default dict is
            training_params = {
            'adam_lr': 0.001,
            'adam_lr_decay_rate': 1000,
            'steps': 2000,
            'adam_b1': 0.9,
            'adam_b2': 0.999}
        jit (bool): True for just-in-time compile.
        fwd_ad(bool): True for using forward-mode auto diff.
    """
    # initialize hyper parameters
    training_params = {
        'adam_lr': 0.001,
        'adam_lr_decay_rate': 1000,
        'steps': 2000,
        'adam_b1': 0.9,
        'adam_b2': 0.999
    }
    training_params.update(options)
    print(training_params)

    lr = training_params['adam_lr']

    def schedual_lr(step):
        return training_params['adam_lr'] * 2**(
            -step / training_params['adam_lr_decay_rate'])

    opt_init, opt_update, get_params = optimizers.adam(
        schedual_lr,
        b1=training_params['adam_b1'],
        b2=training_params['adam_b2'])
    opt_state = opt_init(x0)
    if fwd_ad:
        grad_obj = value_and_gradfwd(fun)
    else:
        grad_obj = jax.value_and_grad(fun)
    if jit:
        grad_obj = jax.jit(grad_obj)

    def step(step, opt_state):

        def grad_clip(p, g):
            return np.clip(g, -0.001 * np.abs(p) / lr, 0.001 * np.abs(p) / lr)

        value, grads = grad_obj(get_params(opt_state), *args)
        grads = jax.tree_util.tree_map(grad_clip, get_params(opt_state), grads)
        pp = pprint.PrettyPrinter(indent=2)
        print('gradient:')
        pp.pprint(convert_to_json_compatible(grads))
        opt_state = opt_update(step, grads, opt_state)
        return value, opt_state

    for i in range(training_params['steps']):
        pp = pprint.PrettyPrinter(indent=2)
        print(f'step: {i}')
        print('parameters:')
        pp.pprint(convert_to_json_compatible(get_params(opt_state)))
        value, opt_state = step(i, opt_state)
        print(f'loss: {value}')

    return get_params(opt_state)



from typing import Tuple
import itertools
import numpy as np
import jax
import jax.numpy as jnp
import jaxopt
import haiku as hk
from numpy import ndarray
from rich import print as rprint
import matplotlib.pyplot as plt

import supergrad
from supergrad.scgraph.graph import (parse_pre_comp_name, parse_post_comp_name)
from .optimize import scipy_minimize

pauli_mats = []
pauli_mats.append(np.eye(2))
pauli_mats.append(np.array([[0., 1.], [1., 0.]]))
pauli_mats.append(np.array([[0., 1j], [-1j, 0.]]))
pauli_mats.append(np.array(np.diag((1., -1.))))


def compute_average_fidelity_with_leakage(u_target, u_computed):
    """
    Compute fidelity with auto state average

    Args:
        u_target (array): Target unitary
        u_computed (array): Test unitary
    """

    n = u_computed.shape[0]
    res = (jnp.trace(jnp.conj(u_computed.T) @ u_computed) +
           abs(jnp.trace(jnp.conj(u_target.T) @ u_computed))**2) / n / (n + 1)
    return jnp.abs(res)


def compute_state_fidelity(psi_target, psi_computed):
    """
    Compute state fidelity

    Args:
        psi_target (array): Target state
        psi_computed (array): Test state
    """
    return jnp.abs(jnp.vdot(psi_target, psi_computed))**2


def apply_2nz(ps, u):
    """Applies n Z-gate before U and other n after U.
    Each single qubit rotation is reorganized as on axis theta.

    Args:
        ps: 2n z_j coefficients for single qubit rotation
        u (ndarray): test unitary

    Returns:
        ndarray: Unitary with single qubit rotations applied
    """
    # Single qubit gate in 1Q system
    n_qubit = round(len(ps) / 2)
    pauli_0 = np.ones(2)
    pauli_3 = np.array([1, -1])
    pre_unitaries = [
        jnp.cos(pre_params) * pauli_0 + 1j * jnp.sin(pre_params) * pauli_3
        for pre_params in ps[:n_qubit]
    ]
    post_unitaries = [
        jnp.cos(post_params) * pauli_0 + 1j * jnp.sin(post_params) * pauli_3
        for post_params in ps[n_qubit:]
    ]
    u2 = supergrad.tensor(*pre_unitaries)
    u3 = supergrad.tensor(*post_unitaries)
    return (u3 * u.T).T * u2


def apply_6nsq_axis(ps, u):
    """
    Apply 3n single qubit rotation before U and other 3n after U
    Each 3 single qubit rotation are reorganized as on axis theta, phi and
    angle gamma.

    Args:
        ps (array): 6n parameters for single qubit rotation
        u (array): unitary
    """

    # Single qubit gate in 1Q system
    n_qubit = round(len(ps) / 2)

    pre_unitaries = [
        jnp.cos(pre_params[1]) * pauli_mats[0] + 1j * jnp.sin(pre_params[1]) *
        (jnp.cos(pre_params[0]) * pauli_mats[3] + jnp.sin(pre_params[0]) *
         (jnp.cos(pre_params[2]) * pauli_mats[1] +
          jnp.sin(pre_params[2]) * pauli_mats[2]))
        for pre_params in ps[:n_qubit]
    ]
    post_unitaries = [
        jnp.cos(post_params[1]) * pauli_mats[0] + 1j * jnp.sin(post_params[1]) *
        (jnp.cos(post_params[0]) * pauli_mats[3] + jnp.sin(post_params[0]) *
         (jnp.cos(post_params[2]) * pauli_mats[1] +
          jnp.sin(post_params[2]) * pauli_mats[2]))
        for post_params in ps[n_qubit:]
    ]

    u2 = supergrad.tensor(*pre_unitaries)
    u3 = supergrad.tensor(*post_unitaries)
    return u3 @ u @ u2


def apply_nz(params, psi):
    """
    Apply n Z-gate after finial state.

    Args:
        params (array): n parameters for single qubit rotation
        psi (array): state
    """
    post_diag = [
        jnp.array([
            jnp.cos(post_params) + 1j * jnp.sin(post_params),
            jnp.cos(post_params) - 1j * jnp.sin(post_params)
        ]) for post_params in params
    ]

    ud3 = supergrad.tensor(*post_diag)
    return ud3[:, jnp.newaxis] * psi


def apply_3nsq_axis(params, psi):
    """
    Apply 3n single qubit rotation after final state.

    Args:
        params (array): 3n parameters for single qubit rotation
        psi (array): state
    """
    post_unitaries = [
        jnp.cos(post_params[1]) * pauli_mats[0] + 1j * jnp.sin(post_params[1]) *
        (jnp.cos(post_params[0]) * pauli_mats[3] + jnp.sin(post_params[0]) *
         (jnp.cos(post_params[2]) * pauli_mats[1] +
          jnp.sin(post_params[2]) * pauli_mats[2])) for post_params in params
    ]

    u3 = supergrad.tensor(*post_unitaries)
    return u3 @ psi


def log_infidelity_with_vz(params, u_target, u_computed):
    """
    Compute infidelity loss with virtual Z gates.

    Args:
        params (array): phase compensation parameters
        u_target (array): Target unitary
        u_computed (array): Unitary
    """
    u_target = jnp.asarray(u_target)
    u_computed = jnp.asarray(u_computed)
    u2 = apply_2nz(params, u_computed)
    return jnp.log(
        (1 - compute_average_fidelity_with_leakage(u_target, u2).real))


def log_infidelity_with_vsq(params, u_target, u_computed):
    """
    Compute infidelity loss with virtual single qubit gates.

    Args:
        params (array): phase compensation parameters
        u_target (array): Target unitary
        u_computed (array): Unitary
    """
    u_target = jnp.asarray(u_target)
    u_computed = jnp.asarray(u_computed)
    u2 = apply_6nsq_axis(params, u_computed)
    return jnp.log(
        (1 - compute_average_fidelity_with_leakage(u_target, u2).real))


def compute_compensation_iswap_gate_vz(u_target, u_computed):
    """Iswap gate with phase compensation. Performing the compensation through
    virtual Z gates.

    Args:
        u_target (array): Target unitary
        u_computed (array): Unitary
    """
    # check u_target is iswap gate
    # assert jnp.sum((iswap().full() - u_target)**
    #                2) < 1e-4, 'The compensation only support iswap gate'
    # compensate U with a diagonal matrix D
    phase00 = u_computed[0, 0] / jnp.abs(u_computed[0, 0])
    phase12 = u_computed[1, 2] / jnp.abs(u_computed[1, 2])
    phase21 = u_computed[2, 1] / jnp.abs(u_computed[2, 1])
    u_computed = u_computed @ jnp.conj(
        jnp.diag(
            jnp.array([
                phase00, -phase21 * 1j, -phase12 * 1j,
                -phase12 * phase21 / phase00
            ])))

    return u_computed


def simple_vz_compensation_angle(u_target,
                                 u_computed) -> Tuple[np.ndarray, jnp.ndarray]:
    """Gets an estimated single sided compensation VZ angle.
    This follows u_target = D @ u_computed,
    so should be applied to cases where u_computed is almost a monomial.
    The order D @ u_computed is chosen because for CR gate most of the
    single qubit phase is from the rotating frame.
    Also assume qubits.

    Args:
        u_target (array): Target unitary
        u_computed (array): Unitary

    Returns:
        an array of angle per qubit to rotate
        a unitary matrix used as (U*U_computed) to get the compensated unitary
    """

    diff = u_target @ u_computed.conj().T
    d = jnp.diag(diff)
    n = int(np.log2(d.size))
    single_q_phase_lst = []
    for i in range(n):
        single_q_phase_lst.append(d[0] / d[2**(n - 1 - i)])
        single_q_phase_lst[-1] = single_q_phase_lst[-1] / jnp.abs(
            single_q_phase_lst[-1])

    vz_diag = jnp.array([
        1,
    ])
    for i in range(n):
        vz_diag = jnp.kron(vz_diag, jnp.array([1, single_q_phase_lst[i]**(-1)]))

    return np.angle(np.array(single_q_phase_lst)), jnp.diag(vz_diag)


def simple_vz_compensation(u_target, u_computed) -> jnp.ndarray:
    """This is currently a single sided compensation,
    i.e. u_target = D @ u_computed,
    so should be applied to cases where u_computed is almost a monomial.
    The order D @ u_computed is chosen because for CR gate most of the
    single qubit phase is from the rotating frame.
    Also assume qubits.

    Args:
        u_target (array): Target unitary
        u_computed (array): Unitary

    Returns:
        the compensated unitary
    """
    return simple_vz_compensation_angle(u_target, u_computed)[1] @ u_computed


def estimate_vz(
    evo,
    params,
    unitary,
    num_qubit,
    transform_matrix=None,
    options: dict = None,
):
    """Estimates the Virtual-Z gate to correct

    Args:
        evo: class `supergrad.Evolve`
        params: params without "single_q_compensation"
        unitary: target unitary
        num_qubit: number of qubits
        transform_matrix: transform matrix to eigen basis
        options: additional keyword arguments to be passed to the `jaxopt`
            optimizer.

    Returns:
        a dictionary of parameters with
    """
    u_computed = evo.eigen_basis(
        hk.data_structures.merge(evo.all_params, params), transform_matrix)

    if options is None:
        options = {}
    error = log_infidelity_with_vz
    initial_guess = jnp.zeros([2 * num_qubit])
    solver = jaxopt.GradientDescent(error, **options)
    res = solver.run(initial_guess, u_target=unitary, u_computed=u_computed)
    print(res.state)
    qubit_list = getattr(evo.graph, 'qubit_subsystem', evo.graph.sorted_nodes)
    list_key = [parse_pre_comp_name(node) for node in qubit_list
               ] + [parse_post_comp_name(node) for node in qubit_list]

    params["single_q_compensation"] = dict([
        (key, val) for key, val in zip(list_key, res.params)
    ])

    return params


def estimate_vsq(
    evo,
    params,
    unitary,
    num_qubit,
    transform_matrix=None,
    initial_guess=None,
    options: dict = None,
    use_scipy_optimizer=False,
):
    """Estimates the virtual single qubit gate to correct.

    Args:
        evo: class `supergrad.Evolve`
        params: params without "single_q_compensation"
        unitary: target unitary
        num_qubit: number of qubits
        transform_matrix: transform matrix to eigen basis
        initial_guess: initial guess for the optimizer
        options: additional keyword arguments to be passed to the `jaxopt`
            optimizer.
        use_scipy_optimizer: use scipy optimizer instead of jaxopt

    Returns:
        a dictionary of parameters with
    """
    u_computed = evo.eigen_basis(
        hk.data_structures.merge(evo.all_params, params), transform_matrix)

    if options is None:
        options = {}
    error = log_infidelity_with_vsq

    if initial_guess is None:
        initial_guess = jnp.zeros([2 * num_qubit, 3])
    elif isinstance(initial_guess.get('single_q_compensation', None), dict):
        initial_guess = jnp.array(
            [val for val in initial_guess['single_q_compensation'].values()])
    if not use_scipy_optimizer:
        solver = jaxopt.GradientDescent(error, **options)
        res = solver.run(initial_guess, u_target=unitary, u_computed=u_computed)
        print(res.state)
    else:
        # create inline class with attribute 'params'
        class res(object):
            pass

        res.params = scipy_minimize(error,
                                    initial_guess,
                                    args=(unitary, u_computed),
                                    logging=True,
                                    **options)['x']
    qubit_list = getattr(evo.graph, 'qubit_subsystem', evo.graph.sorted_nodes)
    list_key = [parse_pre_comp_name(node) for node in qubit_list
               ] + [parse_post_comp_name(node) for node in qubit_list]

    params["single_q_compensation"] = dict([
        (key, val) for key, val in zip(list_key, res.params)
    ])

    return params


def compute_overlap_with_1q_rotation_axis(psi_target,
                                          psi_computed,
                                          opt_method: str = 'gd',
                                          compensation_option='only_vz',
                                          options: dict = None):
    """
    Compute inner product of states, allows for any single qubit rotation after

    Args:
        psi_target (array): Target state
        psi_computed (array): State
        opt_method (string): optimizer, set `gd` for using the Gradient Descent
            solver; set `lbfgs` for using the LBFGS solver, set `iswap_vz` for
            performing the iswap gate compensation through virtual Z gates.
        compensation_option: (string)
            Set single qubit compensation strategy, should be in
            ['no_comp', 'only_vz', 'arbit_single']
        options (dict): additional keyword arguments to be passed to the `jaxopt`
            optimizer.
    """
    assert compensation_option in ['no_comp', 'only_vz', 'arbit_single']
    n_qubit = round(np.log2(psi_target.shape[0]))
    if options is None:
        options = {}

    # initialize single qubit compensation
    if compensation_option == 'only_vz':

        def error(params, psi_target, psi_computed):
            psi_compensated = apply_nz(params, psi_computed)
            return 1 - compute_state_fidelity(psi_target, psi_compensated)

        initial_guess = jnp.zeros([n_qubit])
    elif compensation_option == 'arbit_single':

        def error(params, psi_target, psi_computed):
            psi_compensated = apply_3nsq_axis(params, psi_computed)
            return 1 - compute_state_fidelity(psi_target, psi_compensated)

        initial_guess = jnp.zeros([n_qubit, 3])
    elif compensation_option == 'no_comp':
        return compute_state_fidelity(psi_target, psi_computed), psi_computed

    if opt_method == 'lbfgs':
        # call jaxopt differentiable LBFGS solver
        solver = jaxopt.LBFGS(error, **options)
    elif opt_method == 'gd':
        solver = jaxopt.GradientDescent(error, **options)
    elif opt_method is None:
        return compute_state_fidelity(psi_target, psi_computed), psi_computed
    else:
        raise ValueError('opt_method should be `gd`, `lbfgs`.')

    res = solver.run(initial_guess,
                     psi_target=psi_target,
                     psi_computed=psi_computed)
    # print(initial_guess)
    # print(res.state)
    # print(res.params)
    if compensation_option == 'only_vz':
        psi_compensated = apply_nz(res.params, psi_computed)
    else:
        psi_compensated = apply_3nsq_axis(res.params, psi_computed)

    return compute_state_fidelity(psi_target, psi_compensated), psi_compensated


def compute_fidelity_with_1q_rotation_axis(u_target,
                                           u_computed,
                                           opt_method: str = 'gd',
                                           compensation_option='only_vz',
                                           options: dict = None):
    """
    Compute fidelity, allows for any single qubit rotation before/after

    Args:
        u_target (array): Target unitary
        u_computed (array): Unitary
        opt_method (string): optimizer, set `gd` for using the Gradient Descent
            solver; set `lbfgs` for using the LBFGS solver, set `iswap_vz` for
            performing the iswap gate compensation through virtual Z gates.
        compensation_option: (string)
            Set single qubit compensation strategy, should be in
            ['no_comp', 'only_vz', 'arbit_single']
        options (dict): additional keyword arguments to be passed to the `jaxopt`
            optimizer.
    """
    assert compensation_option in ['no_comp', 'only_vz', 'arbit_single']
    n_qubit = round(np.log2(u_target.shape[0]))
    if options is None:
        options = {}
    f = compute_average_fidelity_with_leakage
    # initialize single qubit compensation
    if compensation_option == 'only_vz':
        error = log_infidelity_with_vz
        initial_guess = jnp.zeros([2 * n_qubit])
    elif compensation_option == 'arbit_single':
        error = log_infidelity_with_vsq
        initial_guess = jnp.zeros([2 * n_qubit, 3])
    elif compensation_option == 'no_comp':
        return f(u_target, u_computed).real, u_computed

    # Create initial guess from simple_vz
    if opt_method is not None and opt_method.endswith("init2"):
        ar_angle_svz, u_init = simple_vz_compensation_angle(
            u_target, u_computed)
        # Convert to the apply_x format
        # single matrix element phase angle to sigmaz angle (/2)
        # single side to pre and post (/2)
        ar_angle = ar_angle_svz / 4
        if compensation_option == "only_vz":
            initial_guess = jnp.array(np.concatenate([ar_angle, ar_angle]))
        elif compensation_option == 'arbit_single':
            guess_side = np.concatenate(
                [np.zeros(ar_angle.shape),
                 np.zeros(ar_angle.shape), ar_angle])
            initial_guess = jnp.array(np.concatenate([guess_side, guess_side]))

    if opt_method == 'lbfgs' or opt_method == "lbfgs_init2":
        # call jaxopt differentiable LBFGS solver
        solver = jaxopt.LBFGS(error, **options)
    elif opt_method == 'gd' or opt_method == 'gd_init2':
        solver = jaxopt.GradientDescent(error, **options)
    elif opt_method == 'iswap_vz':
        u_optimal = compute_compensation_iswap_gate_vz(u_target, u_computed)
        return f(u_target, u_optimal).real, u_optimal
    elif opt_method == 'simple_vz':
        u_optimal = simple_vz_compensation(u_target, u_computed)
        return f(u_target, u_optimal).real, u_optimal
    elif opt_method is None:
        return f(u_target, u_computed).real, u_computed
    else:
        raise ValueError(
            'opt_method should be `gd`, `gd_init2`, `lbfgs`, `simple_vz` or `iswap_vz`.'
        )

    res = solver.run(initial_guess, u_target=u_target, u_computed=u_computed)
    # print(initial_guess)
    # print(res.state)
    # print(res.params)
    if compensation_option == 'only_vz':
        u_optimal = apply_2nz(res.params, u_computed)
    else:
        u_optimal = apply_6nsq_axis(res.params, u_computed)

    return f(u_target, u_optimal).real, u_optimal


def u_to_pauli(u: ndarray) -> jnp.ndarray:
    """Converts unitary to weights of pauli errors.

    Args:
        u (ndarray): A unitary.
            We expect we already multiply it with the inverse of target unitary.

    Returns:
        jnp.ndarray: a 2D array
    """

    num_qubits = int(np.round(np.log2(u.shape[0])))

    def scan_func(carry, target_idx):
        pauli_mats_array = jnp.array(pauli_mats)
        u_pauli = supergrad.tensor(*pauli_mats_array[target_idx])
        pauli_weights = jnp.abs(jnp.trace(u @ u_pauli))**2
        return carry, pauli_weights

    idxs = jnp.array(
        [idx for idx in itertools.product(range(4), repeat=num_qubits)])
    _, pauli_weights_array = jax.lax.scan(scan_func, 0, idxs)

    pauli_weights_tensor = jnp.reshape(pauli_weights_array, (4,) * num_qubits)

    return pauli_weights_tensor / 4**num_qubits


def cphase_angle_opt(u):
    """Optimizing the cphase angle.

    Args:
        u (array): a unitary
    """

    def loss(angles):
        sq_angle1, sq_angle2, c_angle = angles
        sq_phase1 = jnp.exp(1j * sq_angle1)
        sq_phase2 = jnp.exp(1j * sq_angle2)
        u_compensated = u @ jnp.diag(
            jnp.array([1, sq_phase1, sq_phase2, sq_phase1 * sq_phase2]))
        u_target = jnp.diag(jnp.array([1, 1, 1, jnp.exp(1j * c_angle)]))
        return 1 - compute_average_fidelity_with_leakage(
            u_target, u_compensated)

    res = scipy_minimize(loss, (0., 0., 0.))

    return res


def pauli_diagnose(target_u, sim_u, num_print=10):
    """Pauli error diagnose.

    Args:
        target_u (array): target unitary
        u (array): the unitary from time evolution
        num_print (int): The top num_print pauli error operators will be printed
    """

    def pauli_error_string(error_list):
        return ''.join([pauli_dict[i] for i in error_list])

    pauli_dict = {0: 'I', 1: 'X', 2: 'Y', 3: 'Z'}

    err = sim_u.conj().T @ target_u
    pe = u_to_pauli(err)
    ind = np.unravel_index(np.argsort(pe, axis=None), pe.shape)
    ind_array = np.array(ind)
    for i in range(1, num_print + 1):
        print(pauli_error_string(ind_array[:, -i]), f'{pe[ind][-i]:10.4e}')

    return pe, ind_array


def pauli_diagnose_partition(target_u,
                             sim_u,
                             num_print=10,
                             part_lst=None,
                             default_color='white'):
    """Pauli error diagnose. Color the print result by how many partitions have errors

    Args:
        target_u (array): target unitary
        u (array): the unitary from time evolution
        num_print (int): The top num_print pauli error operators will be printed
        part_lst (list): list of list of int
            each nested list contains all qubit index in special partition
        default_color (string): The default color of error string
    """

    def pauli_error_string(error_list, prob_str):
        count = 0
        for p in part_lst:
            if np.sum(error_list[p]) >= 1:
                count += 1
        err_string = ''.join([pauli_dict[i] for i in error_list])
        err_string = err_string + '   ' + prob_str
        if count < 2:
            err_string = '[' + default_color + ']' + err_string + '[' + default_color + ']'
        if count == 2:
            err_string = '[bright_magenta]' + err_string + '[/bright_magenta]'
        if count == 3:
            err_string = '[bright_red]' + err_string + '[/bright_red]'

        return err_string

    if part_lst is None:
        print('part_lst is None!')
        return pauli_diagnose(target_u, sim_u, num_print)
    pauli_dict = {0: 'I', 1: 'X', 2: 'Y', 3: 'Z'}

    err = sim_u.conj().T @ target_u
    pe = u_to_pauli(err)
    ind = np.unravel_index(np.argsort(pe, axis=None), pe.shape)
    ind_array = np.array(ind)
    for i in range(1, num_print + 1):
        rprint(pauli_error_string(ind_array[:, -i], f'{pe[ind][-i]:10.4e}'))

    return pe, ind_array


def compare_unitary(u1, u2, norm_ord=None):
    """Compare two unitary.

    Args:
        u1 (array): unitary
        u2 (array): unitary
        norm_ord ({non-zero int, inf, -inf, 'fro', 'nuc'}, optional):
            Order of the norm. inf means numpy's `inf` object.
            The default is None.
    Return:
        The maximum value of |u1 - u2|

    """
    # unify matrix phase depend on the maximum of first row
    target_idx = np.argmax(np.abs(u1[0]))
    angle_1 = np.angle(u1[0, target_idx])
    angle_2 = np.angle(u2[0, target_idx])
    # Apply the angle to basis
    u1 = np.exp(-1j * angle_1) * u1
    u2 = np.exp(-1j * angle_2) * u2
    d_unitary = np.abs(u1 - u2)
    # normalize
    d_unitary /= np.linalg.norm(d_unitary, ord=norm_ord)

    return np.max(d_unitary)


def plot_pauli_diagnose(target_u, sim_u, num_print=30):
    """Plot the pauli error diagnose results.

    Args:
        target_u (array): target unitary
        u (array): the unitary from time evolution
        num_print (int): The top num_print pauli error operators will be printed
    """

    def pauli_error_string(error_list):
        return ''.join([pauli_dict[i] for i in error_list])

    pauli_dict = {0: 'I', 1: 'X', 2: 'Y', 3: 'Z'}

    pe, ind_array = pauli_diagnose(target_u, sim_u, num_print)
    ind = np.unravel_index(np.argsort(pe, axis=None), pe.shape)
    pe_lst = [error for error in reversed(pe[ind])]
    ind_lst = [pauli_error_string(idx) for idx in reversed(ind_array.T)]

    plt.scatter(ind_lst[:num_print],
                pe_lst[:num_print],
                label='Standard circuit')
    plt.yscale('log')
    plt.xlabel('Pauli errors')
    plt.ylabel('Coefficients')
    plt.xticks(rotation=90, family='monospace')
    plt.legend()


def plot_pauli_diagnose_partition(target_u,
                                  sim_u,
                                  num_print=30,
                                  part_lst=[[0, 1], [2, 3], [4, 5]],
                                  marker='.',
                                  default_color='black'):
    """Plot the pauli error diagnose results.

    Args:
        target_u (array): target unitary
        u (array): the unitary from time evolution
        num_print (int): The top num_print pauli error operators will be printed
        marker: marker for scatter
        default_color: default color if the index is not assigned in `color_dict`
    """

    def compute_count(error_list):
        count = 0
        if part_lst is not None:
            for p in part_lst:
                if np.sum(error_list[p]) >= 1:
                    count += 1
        return count

    def pauli_error_string(error_list):

        err_string = []
        # construct error list pair
        for p in part_lst:
            str_part = [pauli_dict[key] for key in error_list[p]]
            err_string.append(''.join(str_part))

        return ''.join(err_string)

    pauli_dict = {0: 'I', 1: 'X', 2: 'Y', 3: 'Z'}
    color_dict = {0: 'black', 1: 'teal', 2: 'orange', 3: 'purple', 4: 'red'}
    label_dict = {
        i: r'weight $w(\vec{b}) = $' + f'{i}'
        for i, _ in enumerate(part_lst + [None])
    }

    err = sim_u.conj().T @ target_u
    pe = u_to_pauli(err)
    ind = np.unravel_index(np.argsort(pe, axis=None), pe.shape)
    ind_array = np.array(ind)
    # sort pauli error list
    pe = pe[ind]
    label_flag = np.zeros(len(label_dict))
    for i in range(1, num_print + 1):
        count = compute_count(ind_array[:, -i])
        if count == 0:
            continue  # ignore no error case
        if not label_flag[count]:
            plt.scatter(
                i,  # pauli_error_string(ind_array[:, -i]),
                pe[-i],
                label=label_dict[count],
                color=color_dict.get(count, default_color),
                marker=marker,
                s=60)
            label_flag[count] = 1
        else:
            plt.scatter(
                i,  # pauli_error_string(ind_array[:, -i]),
                pe[-i],
                color=color_dict.get(count, default_color),
                marker=marker,
                s=60)

    plt.yscale('log')
    plt.xlabel('Pauli errors $E$')
    plt.ylabel('Error rates')
    plt.xticks([])
    # plt.legend()



import numpy as np
import jax.scipy.linalg as jla
import pytest

from supergrad.quantum_system.kronobj import KronObj

rng = np.random.default_rng(seed=42)
dims = [2, 4, 3, 5, 2]
locs_list = [[[2]], [[2, 3]], [[4, 1]], [[2, 4, 0]], [[0], [1, 2], [3, 1, 2]]]
conjugate_list = ['dag', 'transpose', 'conjugate']


def random_complex_hermitian(dim):
    a = rng.random((dim, dim)) + 1j * rng.random((dim, dim))
    return (a + a.conj().T) / 2


@pytest.mark.parametrize('locs', locs_list)
def test_diag_representation(locs):
    mats = []
    for loc in locs:
        dims_cfg = [dims[i] for i in loc]
        mats.append([random_complex_hermitian(dim) for dim in dims_cfg])
    kr = KronObj(mats, dims, locs, _nested_inpt=True)
    diag_kr = kr.diagonalize_operator()
    assert all(diag_kr.diag_status)
    assert np.allclose(kr.full(), diag_kr.full())
    mixed_kr = kr + diag_kr
    assert np.allclose(mixed_kr.full(), 2 * kr.full())


@pytest.mark.parametrize('locs', locs_list)
@pytest.mark.parametrize('cj_type', conjugate_list)
def test_diag_conjugate(locs, cj_type):
    mats = []
    for loc in locs:
        dims_cfg = [dims[i] for i in loc]
        mats.append([random_complex_hermitian(dim) for dim in dims_cfg])
    kr = KronObj(mats, dims, locs, _nested_inpt=True)
    diag_kr = kr.diagonalize_operator()
    mixed_kr = kr + diag_kr
    for i, k in enumerate([diag_kr, mixed_kr], 1):
        if cj_type == 'dag':
            assert np.allclose(np.conj(kr.full() * i).T, k.dag().full())
        elif cj_type == 'transpose':
            assert np.allclose(np.transpose(kr.full() * i),
                               k.transpose().full())
        elif cj_type == 'conjugate':
            assert np.allclose(np.conj(kr.full()) * i, k.conjugate().full())


# Skip the last one with Trotter
@pytest.mark.parametrize('locs', locs_list[:-1])
def test_diag_exp_vecmul(locs):
    mats = []
    for loc in locs:
        dims_cfg = [dims[i] for i in loc]
        mats.append([random_complex_hermitian(dim) for dim in dims_cfg])
    kr = KronObj(mats, dims, locs, _nested_inpt=True)
    diag_kr = kr.diagonalize_operator()
    psi = rng.random(kr.shape[0])[:, np.newaxis]
    assert np.allclose(diag_kr.expm(psi, 1), kr.expm(psi, 1))
    assert np.allclose(kr.expm(psi, 1), jla.expm(kr.full()) @ psi)
    assert np.allclose(diag_kr.expm(psi, 1), jla.expm(diag_kr.full()) @ psi)



from supergrad.helper.helper import Helper
from supergrad.scgraph.graph import SCGraph


class Spectrum(Helper):
    """Helper for constructing spectrum computing function of the quantum system
    based on the graph which contain all information about qubits and pulse.
    The functions constructed by this way are pure and transformable by JAX.

    Args:
        graph (SCGraph): The graph containing all Hamiltonian parameters.
        truncated_dim (int): desired dimension of the truncated qubit subsystem
        add_random (bool): If true, will add random deviations to the device parameters
        share_params (bool): Share device parameters between the qubits that
            have the same shared_param_mark. This is used only for gradient computation.
            One must define `shared_param_mark` in the `graph.nodes['qubit']['shared_param_mark']`.
        unify_coupling (bool): Let all couplings in the quantum system be the same.
            TODO: if set to true, which coupling will be used to do the computation?
    """

    def __init__(self,
                 graph,
                 truncated_dim=5,
                 add_random=True,
                 share_params=False,
                 unify_coupling=False,
                 *args,
                 **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.graph: SCGraph = graph
        self.truncated_dim = truncated_dim
        self.add_random = add_random
        self.share_params = share_params
        self.unify_coupling = unify_coupling

    def _init_quantum_system(self):
        self.hilbertspace = self.graph.convert_graph_to_quantum_system(
            add_random=self.add_random,
            share_params=self.share_params,
            unify_coupling=self.unify_coupling,
            truncated_dim=self.truncated_dim,
            **self.kwargs)

    def energy_tensor(self, greedy_assign=True):
        """Return the eigenenergy of quantum system in tensor form.

        Args:
            greedy_assign (bool): if True, use greedy assignment mode
                The greedy assignment mode ignores the issue "same state be assigned
                multiple times", due to the weak coupling assumption."""

        return self.hilbertspace.compute_energy_map(greedy_assign)

    def get_model_eigen_basis(self,
                              list_qubit_name,
                              list_drive_subsys,
                              greedy_assign=True):
        """Compute the quantum system properties in multi-qubit eigen basis.
        For example, energy tensor, n operator in eigen basis, phi operator and
        transform matrix will be computed in the same pure function.

        Args:
            list_qubit_name: qubit name list of string.
            list_drive_subsys: driving subsystem list of string.
            greedy_assign (bool): if True, use greedy assignment mode
                The greedy assignment mode ignores the issue "same state be assigned
                multiple times", due to the weak coupling assumption.
        """
        energy_nd = self.hilbertspace.compute_energy_map(greedy_assign)
        transform_matrix = self.hilbertspace.compute_transform_matrix()
        list_sub_energy_nd = []
        list_sub_n_mat = []
        list_sub_phi_mat = []
        for drive_subsys in list_drive_subsys:
            list_n_mat = [
                self.hilbertspace.transform_operator(
                    self.hilbertspace[name].n_operator).reshape(self.dims * 2)
                for name in drive_subsys
            ]
            list_phi_mat = [
                self.hilbertspace.transform_operator(
                    self.hilbertspace[name].phi_operator).reshape(self.dims * 2)
                for name in drive_subsys
            ]
            # generate the slice index for drive_subsys, all the qubit not in the
            # drive_subsys will be set to 0
            ix = tuple([
                slice(None) if name in drive_subsys else 0
                for name in list_qubit_name
            ])
            if not any(ix):
                raise ValueError(
                    'The tensor slice index is all 0, please check the input.')
            sub_energy_nd = energy_nd[ix]
            list_sub_energy_nd.append(sub_energy_nd - sub_energy_nd.min())
            list_sub_n_mat.append([x[ix + ix] for x in list_n_mat])
            list_sub_phi_mat.append([x[ix + ix] for x in list_phi_mat])
        return list_sub_energy_nd, list_sub_n_mat, list_sub_phi_mat, transform_matrix



from typing import Callable, Union, List
import re
import numpy as np
import jax
import jax.numpy as jnp
import haiku as hk

from supergrad.utils.utility import identity_wrap, const_init
from .base import QuantumSystem
from .kronobj import KronObj


def parse_interaction(**kwargs):
    """Parse subsystem_list and keyword arguments, specify the interaction
    between subsystems constructing the `InteractionTerm`.
    Here we try to provide a similar input format as the add_interaction function
    of scqubits.

    Examples:

        parse_interaction(strength=<float>,
                          op1=<Callable>,
                          op2=<Callable>,
                          …,
                          add_hc=<bool>,
                          name=<str>)

    Args:
        strength:
            coefficient parametrize the interaction strength,
            set `None` if you want to pass parameters by dm-haiku.
        op* (callable): subsystems operators
        add_hc: Default is False.
            If set to True, the Hermitian conjugate is added.
        name: module name

    """

    g = kwargs.pop("g", None)
    if g is None:
        g = kwargs.pop("strength", None)
    if g is None:
        g = kwargs.pop("g_strength", None)
    if g is None:
        g = kwargs.pop('J', None)

    add_hc = kwargs.pop("add_hc", False)
    constant = kwargs.pop('constant', False)
    name = kwargs.pop('name', 'interaction_term')

    operator_list = []
    for key, val in kwargs.items():
        if re.match(r"op\d+$", key) is None:
            raise TypeError(f"Unexpected keyword argument {key}.")
        operator_list.append(val)

    return InteractionTerm(strength=g,
                           operator_list=operator_list,
                           add_hc=add_hc,
                           constant=constant,
                           name=name)


class InteractionTerm(hk.Module):
    """Class for specifying a term in the interaction Hamiltonian of a composite
    Hilbert space.

    Args:
        operator_list: list of operator
        strength: coefficient for the interaction strength,
            set `None` if you want to pass parameters by dm-haiku.
        add_hc: If set to True, the Hermitian conjugate is added.
        constant (bool): True for manually setting parameters, False for using
            haiku's parameters management.
        name: module name
    """

    def __init__(
        self,
        operator_list: List[Callable],
        strength: float = None,
        add_hc: bool = False,
        constant: bool = False,
        name: str = 'interaction_term'
    ) -> None:
        super().__init__(name=name)
        if not constant:
            self.strength = hk.get_parameter('strength', [],
                                             init=const_init(strength))
        else:
            self.strength = strength
        self.operator_list = operator_list
        self.add_hc = add_hc

    def hamiltonian(
            self,
            subsystem_list: List) -> jnp.ndarray:
        """Returns the full Hamiltonian of the interacting quantum system
        described by the InteractingSystem object

        Args:
            subsystem_list:
                list of all quantum systems in InteractingSystem calling ``hamiltonian``,
                needed for identity wrapping
        """

        hamiltonian = self.id_wrap_all_ops(self.operator_list,
                                           subsystem_list)
        if self.add_hc:
            hamiltonian += hamiltonian.dag()
        return hamiltonian

    def id_wrap_all_ops(
            self,
            operator_list: List[Callable],
            subsystem_list: List[QuantumSystem]) -> KronObj:
        """Construct a KronObj which represents the interaction operator on the
        Hilbert space. Identity operators of other subsystems are tensored to the
        interaction operator
        """

        mat_list = []
        subsys_index = []
        dims = [subsystem.dim for subsystem in subsystem_list]
        for idx, operator in enumerate(operator_list):
            subsys_index.append(subsystem_list.index(operator.__self__))
            if idx == 0:
                coeff = self.strength
            else:
                coeff = 1
            # call operator
            mat_list.append(coeff * operator())

        return KronObj(mat_list, dims, subsys_index)


class InteractingSystem(QuantumSystem):
    """Class holding information about the whole system.

    Args:
        subsystem_list(list): list of `QuantumSystem`.
        interaction_list: list of `InteractionTerm`, one can use
            `parse_interaction` to construct `InteractionTerm` easily.
        name: module name
    """

    def __init__(self,
                 subsystem_list: List[QuantumSystem],
                 interaction_list: List[InteractionTerm] = None,
                 name='hilbert_space') -> None:
        super().__init__(name=name)

        # Initialize subsystem
        self._subsystems: List[QuantumSystem] = tuple(subsystem_list)
        self._subsystem_by_id_str = {
            subsystem.name: subsystem
            for subsystem in self._subsystems}
        # Initialize interaction
        if isinstance(interaction_list, InteractionTerm):
            interaction_list = [interaction_list]
        if interaction_list is None:
            interaction_list = []
        self.interaction_list = interaction_list
        self._interaction_term_by_id_str = {
            interaction_term.name: interaction_term
            for interaction_term in self.interaction_list}
        self._truncated_dim = self.dim

        # Cached variable
        self.eigval = None
        self.eigvec = None
        self.ar_index_map = None

    def __getitem__(self, key: Union[int, str]):

        if isinstance(key, int):
            return self._subsystems[key]
        if key in self._subsystem_by_id_str:
            return self._subsystem_by_id_str[key]

        raise KeyError(
            f"Unrecognized key: {key}. Key must be an integer index or a "
            "string specifying a subsystem of InteractingSystem.")

    def __iter__(self):
        return iter(self._subsystems)  # [qbt1,qbt2,in1]

    @property
    def subsystem_list(self) -> List[QuantumSystem]:
        return list(self._subsystems)

    @property
    def subsystem_count(self) -> int:
        """Returns number of subsys_list composing the joint Hilbert space"""
        return len(self._subsystems)

    def idling_hamiltonian_in_prod_basis(self) -> KronObj:
        """Generate hamiltonian in the product basis.

        Return:
            Hamiltonian of the composite system, including the interaction
            between components
        """
        hamiltonian = self.subsystem_hamiltonian()
        hamiltonian += self.coupling_hamiltonian()
        return hamiltonian

    def idling_hamiltonian(self) -> jnp.ndarray:
        """Generate hamiltonian in the product basis.

        Return:
            Hamiltonian of the composite system, including the interaction
            between components
        """
        return self.idling_hamiltonian_in_prod_basis().full()

    def subsystem_hamiltonian(self) -> KronObj:
        """Calculate bare hamiltonian.

        Return:
            composite Hamiltonian composed of bare Hamiltonians of subsys_list
            independent of the external parameter.
        """

        ham_list = []
        for subsys_index, subsys in enumerate(self):
            evals = subsys.eigenenergies()
            ham_list.append(
                KronObj([jnp.diag(evals)], self.dim, [subsys_index]))
        return sum(ham_list)

    def coupling_hamiltonian(self) -> jnp.ndarray:
        """Returns the interaction Hamiltonian, based on the interaction terms
        specified for the current InteractingSystem object.

        Return:
            interaction Hamiltonian
        """

        if not self.interaction_list:
            return 0

        operator_list = []
        for term in self.interaction_list:
            if isinstance(term, InteractionTerm):
                operator_list.append(
                    term.hamiltonian(self.subsystem_list))
            else:
                raise TypeError("Expected an instance of InteractionTerm, "
                                f"got {type(term)} instead.")
        hamiltonian = sum(operator_list)

        return hamiltonian

    def eigenenergies(self, evals_count: int = 6):
        """Returns array of eigenvalues.

        Args:
            evals_count:
                number of desired eigenvalues (default value = 6)
        """
        self.eigval, self.eigvec = self._calc_eigsys()

        return self.eigval[:evals_count]

    @property
    def dim(self):
        """Return Hilbert Space dimension."""
        return tuple([subsystem.dim for subsystem in self])

    @property
    def truncated_dim(self):
        """Return the dimension for truncated eigen basis."""
        return self._truncated_dim

    @truncated_dim.setter
    def truncated_dim(self, value):
        """Overwrite the dimension for truncated eigen basis."""
        if len(self.dim) != len(value):
            raise ValueError(
                'Truncated_dim must be tuple that have the same length'
                'as dim.'
            )
        self._truncated_dim = tuple(value)

    def compute_energy_map(self, greedy_assign=True) -> jnp.ndarray:
        """
        Compute the energy of the whole system in the dressed indices,
        enumerating eigenenergies and eigenstates as j=0,1,2,... starting from
        the ground state.
        Assign results to map back to bare product state indices, for a Hilbert
        space composed of three subsystems TODO, specify tuples of the form (l,m,n)
        where l, m, n denote the excitation levels of each bare subsystem.

        Noted we should calculate the matrices that permute the order of eigenvectors.
        The eigenproblem solver will sort the eigenvectors array by the
        eigenvalues, and one could get the "ascending" order. But we want to
        define a "tensor" order and make the eigenvectors array close to
        identity matrix(just like the coordinate array).

        Args:
            greedy_assign(bool): if True, use greedy assignment mode
                TODO The greedy assignment mode ignores the issue "same state be assigned
                multiple times", due to the weak coupling of subsystems.
        """

        self.eigval, self.eigvec = self._calc_eigsys()
        # Choose the row map to [i, j, ...] state, and find the maximum
        # amplitude the column contain the amplitude is the related eigenvector.
        # Now just use an aggressive greedy search
        amp = (jnp.conj(self.eigvec) * self.eigvec).real
        if not greedy_assign:
            # Initialize mapper matrix
            ar_index_map = jnp.zeros(self.truncated_dim, dtype=int)

            def scan_func(carry, target_idx):
                index_map, amp = carry
                ix = jnp.ravel_multi_index(target_idx, self.dim, 'clip')
                ix2 = jnp.argmax(amp[ix, :])
                # Set the weight of basis ix2 of all wave functions to be 0
                # to avoid assigning the same state for multiple times.
                # This will happen when, for instance:
                # |a> = |0> + |1>, |b> = |0> - |1>
                next_amp = amp.at[:, ix2].set(0)
                next_index_map = index_map.at[tuple(target_idx)].set(ix2)
                carry = [next_index_map, next_amp]
                return carry, None

            idxs = jnp.array([idx for idx in np.ndindex(self.truncated_dim)])
            carry = [ar_index_map, amp]
            (self.ar_index_map, _), _ = jax.lax.scan(scan_func, carry, idxs)

        else:
            # The greedy assignment mode ignores the issue "same state be assigned multiple
            # times", due to the weak coupling of subsystems.
            self.ar_index_map = jnp.argmax(amp, axis=1).reshape(self.truncated_dim)

        energy_nd = self.eigval[self.ar_index_map]

        return energy_nd

    def compute_transform_matrix(self):
        """Computes the matrices related to basis transformation.
        Require `compute_energy_map` to be run first to compute
        the transformation.

        Note all transformation keep the num_basis_resonator.

        Return:
            Unitary transform product basis to eigen basis
        """

        if self.ar_index_map is None:
            self.compute_energy_map()
        # Create quantities in the new Hamiltonian
        # Create a new unitary transform
        dim = np.prod(np.array(self.dim))  # use np for `jit`
        u_map = jnp.eye(dim)[self.ar_index_map.flatten()].T
        # rotate the peak of eigen states to diagonal
        u_product_to_eigen = self.eigvec @ u_map
        # here we must unify phase of eigen states, because eigh return states
        # with arbitrary phase.
        angle = jnp.angle(jnp.diag(u_product_to_eigen))
        # Apply the angle to basis
        ar_phase_sys = jnp.exp(-1j * angle)
        u_product_to_eigen *= ar_phase_sys  # numpy broadcast

        return u_product_to_eigen

    def transform_operator(self, operator, **kwargs):
        """Transform operator to current eigenbasis and truncated eigenbasis.

        Args:
            operator(callable): subsystem operator
            kwargs: keyword arguments pass to callable operator.
        Returns:
            TODO only return 1 thing; `operator in eigenbasis`, `operator in truncated eigenbasis`
        """

        opt_tensor = identity_wrap(operator, self.subsystem_list,
                                   **kwargs)
        opt_tensor = opt_tensor.full()
        u_product_to_eigen = self.compute_transform_matrix()
        opt_eig = jnp.conj(u_product_to_eigen).T @ opt_tensor @ u_product_to_eigen

        return opt_eig



from supergrad.utils.fidelity import compute_fidelity_with_1q_rotation_axis
from supergrad.utils.utility import (tensor, compute_average_photon,
                                     identity_wrap, const_init,
                                     create_state_init, basis, permute)

__all__ = [
    'compute_fidelity_with_1q_rotation_axis', 'tensor',
    'compute_average_photon', 'identity_wrap', 'const_init',
    'create_state_init', 'basis', 'permute'
]



import copy
import numpy as np
import json
from jax.tree_util import tree_leaves, tree_map

import supergrad
from supergrad.scgraph.graph import SCGraph
from supergrad.utils.utility import (convert_to_json_compatible,
                                     convert_to_haiku_dict)
from .data.data_graph import CNOTGatePeriodicGraph

graph = CNOTGatePeriodicGraph(seed=1)
qubit_subset = graph.subgraph(['q02', 'q03', 'q12', 'q13', 'q22', 'q23'])


def graph_to_quantum_system(graph: SCGraph = graph,
                            add_random=False,
                            share_params=False):

    class TestGraph(supergrad.Helper):

        def _init_quantum_system(self):
            self.hilbertspace = graph.convert_graph_to_quantum_system(
                truncated_dim=3,
                phiext=0.5 * 2 * np.pi,
                phi_max=5 * np.pi,
                add_random=add_random,
                share_params=share_params)
            self.hamiltonian_component_and_pulseshape, self.pulse_endtime = graph.convert_graph_to_pulse_lst(
                self.hilbertspace, modulate_wave=True)

        def get_device_parameters(self):
            data = []
            for subsystem in self.hilbertspace.subsystem_list:
                data.append({
                    subsystem.name: [subsystem.ec, subsystem.ej, subsystem.el]
                })
            for interaction in self.hilbertspace.interaction_list:
                data.append({interaction.name: interaction.strength})
            return data

    tg = TestGraph()
    params = graph.convert_graph_to_parameters_haiku(share_params)
    return params, tg.get_device_parameters(params)


def test_sharing_parameters():
    hk_0, params_0 = graph_to_quantum_system(share_params=False)
    hk_1, params_1 = graph_to_quantum_system(share_params=True)
    assert len(hk_0) - len(hk_1) == 100
    assert np.allclose(tree_leaves(params_0), tree_leaves(params_1))


def test_enable_deviations():
    hk_0, params_0 = graph_to_quantum_system(add_random=False)
    hk_1, params_1 = graph_to_quantum_system(add_random=True,
                                             share_params=False)
    hk_2, params_2 = graph_to_quantum_system(add_random=True, share_params=True)
    assert len(hk_1) - len(hk_2) == 100
    assert len(hk_0) == len(hk_1)
    assert np.allclose(tree_leaves(params_1), tree_leaves(params_2))
    assert np.allclose(tree_leaves(params_0), tree_leaves(params_2), rtol=0.05)


def test_update_params(share_params=False):
    qubit_subset_temp = copy.deepcopy(qubit_subset)
    all_params = qubit_subset_temp.convert_graph_to_parameters_haiku(
        share_params=share_params)
    # save parameters to json after optimizing
    save_params = json.dumps(convert_to_json_compatible(all_params))
    # one can analyze and edit parameters
    # load optimized parameters(in json)
    load_params = convert_to_haiku_dict(json.loads(save_params))
    # convert to imaginary number
    load_params = tree_map(lambda x: x * 1.0j, load_params)
    # update graph parameters
    old_qubit_subset = copy.deepcopy(qubit_subset_temp)
    qubit_subset_temp.update_params(load_params, share_params=share_params)
    hk_0, params_0 = graph_to_quantum_system(old_qubit_subset)
    hk_1, params_1 = graph_to_quantum_system(qubit_subset_temp)
    assert np.allclose(tree_leaves(hk_0),
                       tree_leaves(tree_map(lambda x: x * -1.0j, hk_1)))
    assert np.allclose(tree_leaves(params_0),
                       tree_map(lambda x: x * -1.0j, tree_leaves(params_1)))


def test_update_params_sharing_params():
    test_update_params(share_params=True)



from supergrad.time_evolution.schrodinger_solver import (
    sesolve, sesolve_final_states_w_basis_trans)
from supergrad.time_evolution.lindblad_solver import (
    mesolve, mesolve_final_states_w_basis_trans)
from supergrad.time_evolution.ode import ode_expm, _parse_hamiltonian, EvoElement

__all__ = [
    'sesolve', 'mesolve', 'sesolve_final_states_w_basis_trans',
    'mesolve_final_states_w_basis_trans', 'ode_expm', '_parse_hamiltonian',
    'EvoElement'
]



import numpy as np
from pylatex import Document, Package, MultiColumn, Tabular, TextColor
from pylatex.utils import NoEscape

from supergrad.scgraph.graph import SCGraph, _parse_pulse_name, _parse_edges_name
from supergrad.utils.fidelity import u_to_pauli


def generate_params_tabu(graph: SCGraph,
                         add_random=True,
                         gradient=None,
                         unify_coupling=True):
    """Generate latex source codes about parameters table."""

    geometry_options = {
        "head": "40pt",
        "margin": "0.5in",
        "bottom": "0.6in",
        "includeheadfoot": True
    }
    doc = Document(geometry_options=geometry_options)
    doc.packages.append(Package('amsmath'))

    qubit_type = {0: '3', 1: '4', 2: '5', 3: '1', 4: '2', 5: '3'}
    qubit_name = {
        0: '$\mathrm{Q}_{02}$',
        1: '$\mathrm{Q}_{03}$',
        2: '$\mathrm{Q}_{12}$',
        3: '$\mathrm{Q}_{13}$',
        4: '$\mathrm{Q}_{22}$',
        5: '$\mathrm{Q}_{23}$'
    }

    # Add statement table
    with doc.create(Tabular("| l | r | r || l | l | r |",
                            row_height=1.5)) as data_table:
        data_table.add_hline()
        data_table.add_row(
            MultiColumn(3, align='|c||', data='Device parameters'),
            MultiColumn(3, align='c|', data='Control parameters'))
        data_table.add_hline()
        data_table.add_row([
            "Parameter", "Value (GHz)", "Gradient", "Parameter", "Value",
            "Gradient"
        ])
        data_table.add_hline()
        # collect device paramters
        dev_params = []
        nodes_order = graph.sorted_nodes
        for i, node in enumerate(nodes_order):
            for par_name in ['ec', 'ej', 'el']:
                data = graph.nodes[node][par_name]
                if add_random:
                    var = graph.nodes[node].get('variance', None)
                    data *= var.get('ec', 1.0)
                if gradient is not None:
                    # unpack gradient data
                    data_grad = gradient[0][node][par_name]
                    str_grad = f'{data_grad / 2 / np.pi:.3e}'
                else:
                    str_grad = ''
                row = [
                    NoEscape('$E_{' + f'{par_name[-1].upper()}, ' +
                             qubit_type[i] + '}$'), f'{data / 2 / np.pi:.3e}',
                    str_grad
                ]
                dev_params.append(row)
        # collect coupling params
        flag_list = []
        for edge in graph.sorted_edges:
            attr = graph.edges[edge]
            q1, q2 = sorted(edge)
            for k, v in dict(attr).items():
                if k in flag_list and unify_coupling:
                    continue
                flag_list.append(k)
                if unify_coupling:
                    key = _parse_edges_name(k, 'all', 'unify')
                else:
                    key = _parse_edges_name(k, q1, q2)
                data = v['strength']
                if gradient is not None:
                    # unpack gradient data
                    data_grad = gradient[0][key]['strength']
                    str_grad = f'{data_grad / 2 / np.pi:.3e}'
                else:
                    str_grad = ''
                if k == 'capacitive_coupling':
                    type_coup = '$J_{C,'
                elif k == 'inductive_coupling':
                    type_coup = '$J_{L,'
                if unify_coupling:
                    row = [
                        NoEscape(type_coup.split(',')[0] + '}$'),
                        f'{data / 2 / np.pi:.3e}', str_grad
                    ]
                else:
                    row = [
                        NoEscape(
                            type_coup +
                            f'({nodes_order.index(q1)}, {nodes_order.index(q2)})'
                            + '}$'), f'{data / 2 / np.pi:.3e}', str_grad
                    ]
                dev_params.append(row)
        # collect control params
        control_params = []
        for idx, node in enumerate(graph.sorted_nodes):
            data = graph.nodes[node]
            pulse_dict = dict(data).pop('pulse', None)
            if pulse_dict is not None:
                pulse_params = ['amp', 't_ramp', 't_plateau', 'omega_d']
                pulse_name = [
                    '$\epsilon_{{d}', '$t_{\\text{ramp}', '$t_{\\text{plateau}',
                    '$\omega_{{d}'
                ]
                unit_list = ['GHz', 'ns', 'ns', 'GHz']

                # pulse_params = ['amp', 'length', 'omega_d', ]
                # pulse_name = ['$\epsilon_{{d}', '$t_{\\text{single}', '$\omega_{{d}']
                # unit_list = ['GHz', 'ns', 'GHz']

                for i, par_name in enumerate(pulse_params):
                    # unpack gradient data
                    if gradient is not None:
                        pulse = _parse_pulse_name(node, 'pulse',
                                                  pulse_dict['pulse_type'])
                        # unpack gradient data
                        try:
                            data_grad = gradient[1][pulse][par_name]
                        except KeyError:
                            data_grad = gradient[2][pulse][par_name]
                        if par_name in ['omega_d', 'amp']:
                            data_grad = data_grad / np.pi / 2
                        str_grad = f'{data_grad:.3e} '
                    else:
                        str_grad = ''

                    data_par = pulse_dict[par_name]
                    if par_name == 'omega_d':
                        data_par = data_par / np.pi / 2
                        eof = '}/2 \pi$'
                    elif par_name == 'amp':
                        data_par = data_par / np.pi / 2
                        eof = '}$'
                    else:
                        eof = '}$'
                    row = [
                        NoEscape(pulse_name[i] + eof + f'({qubit_name[idx]})'),
                        f'{data_par:.3e} ' + unit_list[i], str_grad
                    ]
                    control_params.append(row)
        # filling
        if len(dev_params) >= len(control_params):
            target_col = dev_params
            fill_col = control_params
        else:
            target_col = control_params
            fill_col = dev_params
        for i, row in enumerate(target_col):
            if i < len(fill_col):
                row.extend(fill_col[i])
                data_table.add_row(row)
            else:
                row.extend(['', '', ''])
                data_table.add_row(row)
        data_table.add_hline()

    doc.generate_pdf("complex_report", clean_tex=False)


def generate_pauli_error_tabu(target_u,
                              sim_u,
                              num_print=10,
                              part_lst=[[0, 1], [2, 3], [4, 5]],
                              default_color='black',
                              average_error=None):
    """Generate latex source codes about pauli error analysis.
    Pauli error diagnose. Color the print result by how many partitions have errors

    Args:
        sim_u: unitary
        target_u: target unitary
        num_print: The top num_print pauli error operators will be printed
    """

    def pauli_error_string(error_list, prob_str):
        count = 0
        if part_lst is not None:
            for p in part_lst:
                if np.sum(error_list[p]) >= 1:
                    count += 1
        if count < 4:
            color = color_dict[count]
        else:
            color = 'blue'

        err_string = []
        # construct error list pair
        for p in part_lst:
            str_part = [pauli_dict[key] for key in error_list[p]]
            err_string.append(TextColor(color, ''.join(str_part)))

        err_string.append(TextColor(color, prob_str))

        return err_string

    pauli_dict = {0: 'I', 1: 'X', 2: 'Y', 3: 'Z'}
    color_dict = {0: default_color, 1: 'teal', 2: 'orange', 3: 'purple'}
    qubit_name = {
        0: '$\mathrm{Q}_{02}$',
        1: '$\mathrm{Q}_{03}$',
        2: '$\mathrm{Q}_{12}$',
        3: '$\mathrm{Q}_{13}$',
        4: '$\mathrm{Q}_{22}$',
        5: '$\mathrm{Q}_{23}$'
    }

    err = sim_u.conj().T @ target_u
    pe = u_to_pauli(err)
    ind = np.unravel_index(np.argsort(pe, axis=None), pe.shape)
    ind_array = np.array(ind)

    geometry_options = {
        "head": "40pt",
        "margin": "0.5in",
        "bottom": "0.6in",
        "includeheadfoot": True
    }
    doc = Document(geometry_options=geometry_options)
    doc.packages.append(Package('amsmath'))

    # Add statement table
    num_part = len(part_lst)
    column_name = [
        NoEscape(' '.join([qubit_name[idx] for idx in p])) for p in part_lst
    ]
    with doc.create(Tabular('| c ' * num_part + '| c |',
                            row_height=1.5)) as data_table:
        data_table.add_hline()
        data_table.add_row(column_name + ['Error Rate'])
        data_table.add_hline()
        for i in range(1, num_print + 1):
            data_table.add_row(
                pauli_error_string(ind_array[:, -i], f'{pe[ind][-i]:10.4e}'))
        data_table.add_hline()
        if average_error is not None:
            data_table.add_hline()
            for i, error in enumerate(average_error, 1):
                color = color_dict[i]
                data_table.add_row([
                    MultiColumn(size=num_part,
                                align='| c |',
                                data=TextColor(
                                    color, f'Average weight-{i} Pauli error')),
                    TextColor(color, f'{error:10.4e}')
                ])
            data_table.add_hline()

    doc.generate_pdf("complex_report", clean_tex=False)



import numpy as np
import jax.numpy as jnp
import haiku as hk

from supergrad.scgraph.graph import (SCGraph, parse_pre_comp_name,
                                     parse_post_comp_name)

pauli_mats = []
pauli_mats.append(np.eye(2))
pauli_mats.append(np.array([[0., 1.], [1., 0.]]))
pauli_mats.append(np.array([[0., 1j], [-1j, 0.]]))
pauli_mats.append(np.array(np.diag((1., -1.))))


class SingleQubitCompensation(hk.Module):
    """Class for single qubit gates compensation.

    Args:
        graph: SuperGrad graph
        compensation_option (string): Set single qubit compensation strategy,
            should be in ['only_vz', 'arbit_single']
        coupler_subsystem: the name of coupler subsystem should be keep in |0>
            during the evolution.
        name: module name
    """

    def __init__(self,
                 graph: SCGraph,
                 compensation_option='only_vz',
                 coupler_subsystem=[],
                 name: str = 'single_q_compensation'):
        super().__init__(name=name)
        assert compensation_option in ['only_vz', 'arbit_single']
        # construct activation function
        self.pre_comp_angles = []
        self.post_comp_angles = []
        self.compensation_option = compensation_option
        shape = [] if self.compensation_option == 'only_vz' else [3]

        for node in graph.sorted_nodes:
            if node not in coupler_subsystem:
                self.pre_comp_angles.append(
                    hk.get_parameter(parse_pre_comp_name(node),
                                     shape,
                                     init=jnp.zeros))
                self.post_comp_angles.append(
                    hk.get_parameter(parse_post_comp_name(node),
                                     shape,
                                     init=jnp.zeros))

    def create_unitaries(self):
        """Create unitaries describe the single-qubit rotation before and after
        the time evolution.
        """
        if self.compensation_option == 'only_vz':
            # do single-qubit Z-rotation before and after the time evolution
            pauli_0 = np.ones(2)
            pauli_3 = np.array([1, -1])
            pre_unitaries = [
                jnp.cos(pre_params) * pauli_0 +
                1j * jnp.sin(pre_params) * pauli_3
                for pre_params in self.pre_comp_angles
            ]
            post_unitaries = [
                jnp.cos(post_params) * pauli_0 +
                1j * jnp.sin(post_params) * pauli_3
                for post_params in self.post_comp_angles
            ]
        else:
            # do arbitrary single-qubit rotation before and after the time evolution
            pre_unitaries = [
                jnp.cos(pre_params[1]) * pauli_mats[0] +
                1j * jnp.sin(pre_params[1]) *
                (jnp.cos(pre_params[0]) * pauli_mats[3] +
                 jnp.sin(pre_params[0]) *
                 (jnp.cos(pre_params[2]) * pauli_mats[1] +
                  jnp.sin(pre_params[2]) * pauli_mats[2]))
                for pre_params in self.pre_comp_angles
            ]
            post_unitaries = [
                jnp.cos(post_params[1]) * pauli_mats[0] +
                1j * jnp.sin(post_params[1]) *
                (jnp.cos(post_params[0]) * pauli_mats[3] +
                 jnp.sin(post_params[0]) *
                 (jnp.cos(post_params[2]) * pauli_mats[1] +
                  jnp.sin(post_params[2]) * pauli_mats[2]))
                for post_params in self.post_comp_angles
            ]

        return pre_unitaries, post_unitaries



"""
The code is from `qutip.operators`.
This module contains functions for generating a variety of commonly occurring
quantum operators.
"""

__all__ = [
    'jmat', 'spin_Jx', 'spin_Jy', 'spin_Jz', 'spin_Jm', 'spin_Jp', 'spin_J_set',
    'sigmap', 'sigmam', 'sigmax', 'sigmay', 'sigmaz', 'destroy', 'create',
    'qeye', 'identity', 'position', 'momentum', 'num', 'squeeze', 'squeezing',
    'displace', 'commutator', 'qutrit_ops', 'qdiags', 'phase', 'qzero',
    'charge', 'tunneling'
]

import numbers
import numpy as np
import jax
import jax.numpy as jnp
import jax.scipy.linalg as jaxLA

from supergrad.utils.docstring import implements
from supergrad.utils.utility import qutrit_basis

try:
    import qutip.core.operators as qop
except ImportError:
    qop = None


#
# Spin operators
#
@implements(getattr(qop, 'jmat', None))
def jmat(j, *args):
    if (np.fix(2 * j) != 2 * j) or (j < 0):
        raise TypeError('j must be a non-negative integer or half-integer')

    if not args:
        return jmat(j, 'x'), jmat(j, 'y'), jmat(j, 'z')

    if args[0] == '+':
        A = _jplus(j)
    elif args[0] == '-':
        A = np.conj(_jplus(j)).T
    elif args[0] == 'x':
        A = 0.5 * (_jplus(j) + np.conj(_jplus(j)).T)
    elif args[0] == 'y':
        A = -0.5 * 1j * (_jplus(j) - np.conj(_jplus(j)).T)
    elif args[0] == 'z':
        A = _jz(j)
    else:
        raise TypeError('Invalid type')

    return A


@implements(getattr(qop, '_jplus', None))
def _jplus(j):
    m = np.arange(j, -j - 1, -1, dtype=complex)
    data = (np.sqrt(j * (j + 1.0) - (m + 1.0) * m))[1:]
    return np.diag(data, k=1)


@implements(getattr(qop, '_jz', None))
def _jz(j):
    n = int(2 * j + 1)
    data = np.array([j - k for k in range(n) if (j - k) != 0], dtype=complex)
    # Even shaped matrix
    if (n % 2 != 0):
        data = np.insert(data, int(j), 0)
    return np.diag(data)


#
# Spin j operators:
#
@implements(getattr(qop, 'spin_Jx', None))
def spin_Jx(j):
    return jmat(j, 'x')


@implements(getattr(qop, 'spin_Jy', None))
def spin_Jy(j):
    return jmat(j, 'y')


@implements(getattr(qop, 'spin_Jz', None))
def spin_Jz(j):
    return jmat(j, 'z')


@implements(getattr(qop, 'spin_Jm', None))
def spin_Jm(j):
    return jmat(j, '-')


@implements(getattr(qop, 'spin_Jp', None))
def spin_Jp(j):
    return jmat(j, '+')


@implements(getattr(qop, 'spin_J_set', None))
def spin_J_set(j):
    return jmat(j)


#
# Pauli spin 1/2 operators:
#
@implements(getattr(qop, 'sigmap', None))
def sigmap():
    return jmat(1 / 2., '+')


@implements(getattr(qop, 'sigmam', None))
def sigmam():
    return jmat(1 / 2., '-')


@implements(getattr(qop, 'sigmax', None))
def sigmax():
    return 2 * jmat(1 / 2, 'x')


@implements(getattr(qop, 'sigmay', None))
def sigmay():
    return 2 * jmat(1 / 2, 'y')


@implements(getattr(qop, 'sigmaz', None))
def sigmaz():
    return 2 * jmat(1 / 2, 'z')


#
# DESTROY returns annihilation operator for n dimensional Hilbert space
# out = destroy(n), n is integer value &  n>0
#
@implements(getattr(qop, 'destroy', None))
def destroy(n, offset=0):
    if not isinstance(n, (int, np.integer)):  # raise error if n not integer
        raise ValueError("Hilbert space dimension must be integer value")
    data = np.sqrt(np.arange(offset + 1, n + offset, dtype=complex))
    return np.diag(data, k=1)


#
# create returns creation operator for n dimensional Hilbert space
# out = create(n), n is integer value &  n>0
#
@implements(getattr(qop, 'create', None))
def create(n, offset=0):
    if not isinstance(n, (int, np.integer)):  # raise error if n not integer
        raise ValueError("Hilbert space dimension must be integer value")
    qo = destroy(n, offset=offset)  # create operator using destroy function
    return np.conj(qo).T


@implements(getattr(qop, '_implicit_tensor_dimensions', None))
def _implicit_tensor_dimensions(dimensions):
    flat = jax.tree_util.tree_leaves(dimensions)
    if not all(isinstance(x, numbers.Integral) and x >= 0 for x in flat):
        raise ValueError("All dimensions must be integers >= 0")
    return np.prod(flat)


@implements(getattr(qop, 'qzero', None))
def qzero(dimensions):
    size = _implicit_tensor_dimensions(dimensions)
    # A sparse matrix with no data is equal to a zero matrix.
    return np.zeros((size, size))


#
# QEYE returns identity operator for a Hilbert space with dimensions dims.
# a = qeye(n), n is integer or list of integers & all elements >= 0
#
@implements(getattr(qop, 'qeye', None))
def qeye(dimensions):
    size = _implicit_tensor_dimensions(dimensions)
    return np.eye(size)


@implements(getattr(qop, 'identity', None))
def identity(dims):
    return qeye(dims)


@implements(getattr(qop, 'position', None))
def position(n, offset=0):
    a = destroy(n, offset=offset)
    return 1.0 / np.sqrt(2.0) * (a + np.conj(a).T)


@implements(getattr(qop, 'momentum', None))
def momentum(n, offset=0):
    a = destroy(n, offset=offset)
    return -1j / np.sqrt(2.0) * (a - np.conj(a).T)


@implements(getattr(qop, 'num', None))
def num(n, offset=0):
    data = np.arange(n, dtype=complex) + offset
    return np.diag(data)


@implements(getattr(qop, 'squeeze', None))
def squeeze(n, z, offset=0):
    a = destroy(n, offset=offset)
    op = (1 / 2.0) * np.conj(z) * (a @ a) - (1 / 2.0) * z * (np.conj(a).T) @ (
        np.conj(a).T)
    return jaxLA.expm(op)


@implements(getattr(qop, 'squeezing', None))
def squeezing(a1, a2, z):
    b = 0.5 * (np.conj(z) * (a1 @ a2) - z * (jnp.conj(a1).T @ jnp.conj(a2).T))
    return jaxLA.expm(b)


@implements(getattr(qop, 'displace', None))
def displace(n, alpha, offset=0):
    a = destroy(n, offset=offset)
    d = (alpha * np.conj(a).T - np.conj(alpha) * a)
    return jaxLA.expm(d)


@implements(getattr(qop, 'commutator', None))
def commutator(A, B, kind="normal"):
    if kind == 'normal':
        return A @ B - B @ A

    elif kind == 'anti':
        return A @ B + B @ A

    else:
        raise TypeError("Unknown commutator kind '%s'" % kind)


@implements(getattr(qop, 'qutrit_ops', None))
def qutrit_ops():
    out = np.empty((6,), dtype=object)
    one, two, three = qutrit_basis()
    out[0] = one * np.conj(one).T
    out[1] = two * np.conf(two).T
    out[2] = three * np.conj(three).T
    out[3] = one * np.conj(two).T
    out[4] = two * np.conj(three).T
    out[5] = three * np.conj(one).T
    return out


@implements(getattr(qop, 'qdiags', None))
def qdiags(diagonals, offsets):
    data = np.diag(diagonals, offsets)
    return data


@implements(getattr(qop, 'phase', None))
def phase(n, phi0=0):
    phim = phi0 + (2.0 * np.pi * np.arange(n)) / n  # discrete phase angles
    n = np.arange(n).reshape((n, 1))
    states = np.array(
        [np.sqrt(kk) / np.sqrt(n) * np.exp(1.0j * n * kk) for kk in phim])
    ops = np.array([np.outer(st, st.conj()) for st in states])
    return np.sum(ops, axis=0)


@implements(getattr(qop, 'charge', None))
def charge(nmax, nmin=None, frac=1):
    if nmin is None:
        nmin = -nmax
    diag = np.arange(nmin, nmax + 1, dtype=float)
    if frac != 1:
        diag *= frac
    return np.diag(diag)


@implements(getattr(qop, 'tunneling', None))
def tunneling(n, m=1):
    diags = np.ones(n - m, dtype=int)
    return np.diag(diags, m) + np.diag(diags, -m)



import numpy as np
import jax
import haiku as hk
import qutip as qt
from qutip import Qobj
from supergrad import KronObj


def to_qutip_operator(superObj: KronObj):
    """Convert the Hamiltonian in `superObj` to qutip format.

    Args:
        superObj: The Hamiltonian in `KronObj` format.
    Returns:
        The Hamiltonian in qutip format.
    """
    # firstly downcast diag unitary
    if any(superObj.diag_status):
        superObj._downcast_diagonal_unitary()
    data = []
    for mats, local_list in zip(superObj.data, superObj.locs):
        if local_list is not None:  # downcasting
            # construct identity matrix list
            mat_list = [Qobj(np.eye(dim)) for dim in superObj.dims]
            for mat, local in zip(mats, local_list):
                mat_list[local] = Qobj(np.array(mat))
            # calculate tensor product
            data.append(qt.tensor(*mat_list))
        else:
            data.append(Qobj(np.array(mats), [
                list(superObj.dims),
            ] * 2))
    return sum(data)


def to_qutip_operator_function_pair(hamiltonian_component_and_pulseshape):
    """Convert the drive Hamiltonian to qutip operator-function pairs.

    Args:
        hamiltonian_component_and_pulseshape: The list of pair containing the
            time-dependent components of the Hamiltonian and corresponding pulse
            shape.
    Returns:
        The list of operator-function pairs.
    """
    operator_function_pair_list = []
    for drive, create_pulse in hamiltonian_component_and_pulseshape:
        oper = to_qutip_operator(drive)
        pulse = create_pulse.__self__

        def pulse_closure(pulse):

            @hk.without_apply_rng
            @hk.transform
            def new_create_pulse(t, args):
                new_pulse = type(pulse)(modulate_wave=True, name=pulse.name)
                return new_pulse.create_pulse(t, args)

            # get parameters template
            rng = jax.random.PRNGKey(0)
            params = new_create_pulse.init(rng, 0., {})
            # filling params by existed pulse
            existed_params = vars(pulse)
            existed_params = hk.data_structures.filter(
                lambda module_name, name, value: name in params[pulse.name].
                keys(), {pulse.name: existed_params})
            params = hk.data_structures.merge(params, existed_params)

            return lambda t, args: float(new_create_pulse.apply(
                params, t, args))

        operator_function_pair_list.append([oper, pulse_closure(pulse)])

    return operator_function_pair_list



from typing import Callable
from abc import ABC, abstractmethod
import jax
import haiku as hk


class Helper(ABC):
    """Transform attribute to pure function.
    """

    def __init__(self, *args, **kwargs) -> None:

        self.args = args
        self.kwargs = kwargs
        self.graph = None
        self.hilbertspace = None

    @abstractmethod
    def _init_quantum_system(self):
        """Initialize quantum system"""

    def __getattribute__(self, attr, transform=True):
        """Use Haiku's decorator before get attribute."""

        origin_attr = super().__getattribute__(attr)
        if transform and isinstance(
                origin_attr,
                Callable) and not (attr.startswith('_') or attr == 'ls_params'):

            @hk.without_apply_rng
            @hk.transform
            def _transform_func(*args, **kwargs):
                self._init_quantum_system()
                return origin_attr(*args, **kwargs)

            return _transform_func.apply
        return origin_attr

    def ls_params(self, attr: str = None, *args, **kwargs):
        """Get all parameters for the selected attribute.

        Args:
            attr(str): Default `None`, get parameters for
            `self._init_quantum_system`.
        """

        @hk.transform
        def prepare_params():
            self._init_quantum_system()
            if attr is not None:
                self.__getattribute__(attr, transform=False)(*args, **kwargs)
            return

        rng = jax.random.PRNGKey(0)
        return prepare_params.init(rng)

    def _get_dims(self):
        """Get the dimension of the composited quantum system."""
        return self.hilbertspace.dim

    def get_dims(self):
        return self._get_dims()

    @property
    def dims(self):
        """The dimension of the composited quantum system.
        Only for internal use temporarily.
        """
        return self._get_dims()

    @property
    def device_params(self):
        """The device parameters dictionary."""
        return self.graph.convert_graph_to_parameters_haiku(
            self.share_params, self.unify_coupling, only_device_params=True)

    @property
    def all_params(self):
        """The static parameters dictionary."""
        all_params = self.graph.convert_graph_to_parameters_haiku(
            self.share_params, self.unify_coupling)
        return all_params

    def idling_hamiltonian_in_prod_basis(self):
        """Return the idling Hamiltonian in product basis."""
        return self.hilbertspace.idling_hamiltonian_in_prod_basis()

    def construct_transform_matrix(self):
        """Constructing the transform matrix from product basis to eigenbasis.
        """
        return self.hilbertspace.compute_transform_matrix()

    def eigensystem(self):
        """Calculate quantum system's eigenenergy and eigenvectors using
        `jax.scipy.linalg.eigh`.
        """
        return self.hilbertspace._calc_eigsys()

    def n_operator(self, node_name, transform=False):
        """Return the operator of the selected node.

        Args:
            node_name (str): name of the selected node.
            transform (bool): if True, return the transformed operator in eigen basis.
        """
        if transform:
            opt = self.hilbertspace.transform_operator(
                self.hilbertspace[node_name].n_operator).reshape(self.dims * 2)
            ix = tuple([
                slice(None) if name == node_name else 0
                for name in self.graph.sorted_nodes
            ])
            if not any(ix):
                raise ValueError(
                    'The tensor slice index is all 0, please check the input.')
            return opt[ix + ix]
        else:
            return self.hilbertspace[node_name].n_operator()

    def phi_operator(self, node_name, transform=False):
        """Return the operator of the selected node.

        Args:
            node_name (str): name of the selected node.
            transform (bool): if True, return the transformed operator in eigen basis.
        """
        if transform:
            opt = self.hilbertspace.transform_operator(
                self.hilbertspace[node_name].phi_operator).reshape(self.dims * 2)
            ix = tuple([
                slice(None) if name == node_name else 0
                for name in self.graph.sorted_nodes
            ])
            if not any(ix):
                raise ValueError(
                    'The tensor slice index is all 0, please check the input.')
            return opt[ix + ix]
        else:
            return self.hilbertspace[node_name].phi_operator()



"""
The code is from `qutip_qip.operations`.
Operations on quantum circuits.
"""
import numbers
from collections.abc import Iterable
import numpy as np
import jax.numpy as jnp
import jax.scipy.linalg as jaxLA

from supergrad.utils.docstring import implements
from supergrad.utils.operators import identity, sigmax, sigmay, sigmaz
from supergrad.utils.utility import tensor, fock_dm, permute

try:
    import qutip_qip.operations as qop
except ImportError:
    qop = None

__all__ = [
    "rx",
    "ry",
    "rz",
    "sqrtnot",
    "snot",
    "phasegate",
    "qrot",
    "x_gate",
    "y_gate",
    "z_gate",
    "cy_gate",
    "cz_gate",
    "s_gate",
    "t_gate",
    "qasmu_gate",
    "cs_gate",
    "ct_gate",
    "cphase",
    "cnot",
    "csign",
    "berkeley",
    "swapalpha",
    "swap",
    "iswap",
    "sqrtswap",
    "sqrtiswap",
    "fredkin",
    "molmer_sorensen",
    "toffoli",
    "rotation",
    "controlled_gate",
    "globalphase",
    "hadamard_transform",
    "expand_operator",
]

#
# Single Qubit Gates
#


@implements(getattr(qop, 'x_gate', None))
def x_gate(n=None, target=0):
    if n is not None:
        return gate_expand_1ton(x_gate(), n, target)
    return sigmax()


@implements(getattr(qop, 'y_gate', None))
def y_gate(n=None, target=0):
    if n is not None:
        return gate_expand_1ton(y_gate(), n, target)
    return sigmay()


@implements(getattr(qop, 'cy_gate', None))
def cy_gate(n=None, control=0, target=1):
    if (control == 1 and target == 0) and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(cy_gate(), n, control, target)
    return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, -1j], [0, 0, 1j, 0]])


@implements(getattr(qop, 'z_gate', None))
def z_gate(n=None, target=0):
    if n is not None:
        return gate_expand_1ton(z_gate(), n, target)
    return sigmaz()


@implements(getattr(qop, 'cz_gate', None))
def cz_gate(n=None, control=0, target=1):
    if (control == 1 and target == 0) and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(cz_gate(), n, control, target)
    return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])


@implements(getattr(qop, 's_gate', None))
def s_gate(n=None, target=0):
    if n is not None:
        return gate_expand_1ton(s_gate(), n, target)
    return np.array([[1, 0], [0, 1j]])


@implements(getattr(qop, 'cs_gate', None))
def cs_gate(n=None, control=0, target=1):
    if (control == 1 and target == 0) and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(cs_gate(), n, control, target)
    return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1j]])


@implements(getattr(qop, 't_gate', None))
def t_gate(n=None, target=0):
    if n is not None:
        return gate_expand_1ton(t_gate(), n, target)
    return np.array([[1, 0], [0, np.exp(1j * np.pi / 4)]])


@implements(getattr(qop, 'ct_gate', None))
def ct_gate(n=None, control=0, target=1):
    if (control == 1 and target == 0) and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(ct_gate(), n, control, target)
    return np.array([
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, np.exp(1j * np.pi / 4)],
    ])


@implements(getattr(qop, 'rx', None))
def rx(phi, n=None, target=0):
    if n is not None:
        return gate_expand_1ton(rx(phi), n, target)
    return jnp.array([
        [jnp.cos(phi / 2), -1j * jnp.sin(phi / 2)],
        [-1j * jnp.sin(phi / 2), jnp.cos(phi / 2)],
    ])


@implements(getattr(qop, 'ry', None))
def ry(phi, n=None, target=0):
    if n is not None:
        return gate_expand_1ton(ry(phi), n, target)
    return jnp.array([
        [jnp.cos(phi / 2), -jnp.sin(phi / 2)],
        [jnp.sin(phi / 2), jnp.cos(phi / 2)],
    ])


@implements(getattr(qop, 'rz', None))
def rz(phi, n=None, target=0):
    if n is not None:
        return gate_expand_1ton(rz(phi), n, target)
    return jnp.array([[jnp.exp(-1j * phi / 2), 0], [0, jnp.exp(1j * phi / 2)]])


@implements(getattr(qop, 'sqrtnot', None))
def sqrtnot(n=None, target=0):
    if n is not None:
        return gate_expand_1ton(sqrtnot(), n, target)
    return np.array([[0.5 + 0.5j, 0.5 - 0.5j], [0.5 - 0.5j, 0.5 + 0.5j]])


@implements(getattr(qop, 'snot', None))
def snot(n=None, target=0):
    if n is not None:
        return gate_expand_1ton(snot(), n, target)
    return 1 / np.sqrt(2.0) * np.array([[1, 1], [1, -1]])


@implements(getattr(qop, 'phasegate', None))
def phasegate(theta, n=None, target=0):
    if n is not None:
        return gate_expand_1ton(phasegate(theta), n, target)
    return jnp.array([[1, 0], [0, jnp.exp(1.0j * theta)]])


@implements(getattr(qop, 'qrot', None))
def qrot(theta, phi, n=None, target=0):
    if n is not None:
        return expand_operator(qrot(theta, phi), n=n, targets=target)
    return jnp.array([
        [
            jnp.cos(theta / 2.0),
            -1.0j * jnp.exp(-1.0j * phi) * jnp.sin(theta / 2.0),
        ],
        [
            -1.0j * jnp.exp(1.0j * phi) * jnp.sin(theta / 2.0),
            jnp.cos(theta / 2.0),
        ],
    ])


@implements(getattr(qop, 'qasmu_gate', None))
def qasmu_gate(args, n=None, target=0):
    theta, phi, gamma = args
    if n is not None:
        return expand_operator(qasmu_gate([theta, phi, gamma]),
                               n=n,
                               targets=target)
    return jnp.array(rz(phi) @ ry(theta) @ rz(gamma))


#
# 2 Qubit Gates
#


@implements(getattr(qop, 'cphase', None))
def cphase(theta, n=2, control=0, target=1):
    if n < 1 or target < 0 or control < 0:
        raise ValueError("Minimum value: n=1, control=0 and target=0")

    if control >= n or target >= n:
        raise ValueError("control and target need to be smaller than N")

    u_list1 = [identity(2)] * n
    u_list2 = [identity(2)] * n

    u_list1[control] = fock_dm(2, 1)
    u_list1[target] = phasegate(theta)

    u_list2[control] = fock_dm(2, 0)

    u = tensor(*u_list1) + tensor(*u_list2)
    return u


@implements(getattr(qop, 'cnot', None))
def cnot(n=None, control=0, target=1):
    if (control == 1 and target == 0) and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(cnot(), n, control, target)
    return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])


@implements(getattr(qop, 'csign', None))
def csign(n=None, control=0, target=1):
    if (control == 1 and target == 0) and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(csign(), n, control, target)
    return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]])


@implements(getattr(qop, 'berkeley', None))
def berkeley(n=None, targets=[0, 1]):
    if (targets[0] == 1 and targets[1] == 0) and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(berkeley(), n, targets=targets)
    return np.array([
        [np.cos(np.pi / 8), 0, 0, 1.0j * np.sin(np.pi / 8)],
        [0, np.cos(3 * np.pi / 8), 1.0j * np.sin(3 * np.pi / 8), 0],
        [0, 1.0j * np.sin(3 * np.pi / 8),
         np.cos(3 * np.pi / 8), 0],
        [1.0j * np.sin(np.pi / 8), 0, 0,
         np.cos(np.pi / 8)],
    ])


@implements(getattr(qop, 'swapalpha', None))
def swapalpha(alpha, n=None, targets=[0, 1]):
    if (targets[0] == 1 and targets[1] == 0) and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(swapalpha(alpha), n, targets=targets)
    return jnp.array([
        [1, 0, 0, 0],
        [
            0,
            0.5 * (1 + jnp.exp(1.0j * jnp.pi * alpha)),
            0.5 * (1 - jnp.exp(1.0j * jnp.pi * alpha)),
            0,
        ],
        [
            0,
            0.5 * (1 - jnp.exp(1.0j * jnp.pi * alpha)),
            0.5 * (1 + jnp.exp(1.0j * jnp.pi * alpha)),
            0,
        ],
        [0, 0, 0, 1],
    ])


@implements(getattr(qop, 'swap', None))
def swap(n=None, targets=[0, 1]):
    if targets != [0, 1] and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(swap(), n, targets=targets)
    return np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])


@implements(getattr(qop, 'iswap', None))
def iswap(n=None, targets=[0, 1]):
    if targets != [0, 1] and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(iswap(), n, targets=targets)
    return np.array([[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]])


@implements(getattr(qop, 'sqrtswap', None))
def sqrtswap(n=None, targets=[0, 1]):
    if targets != [0, 1] and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(sqrtswap(), n, targets=targets)
    return np.array([
        [1, 0, 0, 0],
        [0, 0.5 + 0.5j, 0.5 - 0.5j, 0],
        [0, 0.5 - 0.5j, 0.5 + 0.5j, 0],
        [0, 0, 0, 1],
    ])


@implements(getattr(qop, 'sqrtiswap', None))
def sqrtiswap(n=None, targets=[0, 1]):
    if targets != [0, 1] and n is None:
        n = 2

    if n is not None:
        return gate_expand_2ton(sqrtiswap(), n, targets=targets)
    return np.array([
        [1, 0, 0, 0],
        [0, 1 / np.sqrt(2), 1j / np.sqrt(2), 0],
        [0, 1j / np.sqrt(2), 1 / np.sqrt(2), 0],
        [0, 0, 0, 1],
    ])


@implements(getattr(qop, 'molmer_sorensen', None))
def molmer_sorensen(theta, n=None, targets=[0, 1]):
    if targets != [0, 1] and n is None:
        n = 2

    if n is not None:
        return expand_operator(molmer_sorensen(theta), n, targets=targets)
    return jnp.array([
        [jnp.cos(theta / 2.0), 0, 0, -1.0j * jnp.sin(theta / 2.0)],
        [0, jnp.cos(theta / 2.0), -1.0j * jnp.sin(theta / 2.0), 0],
        [0, -1.0j * jnp.sin(theta / 2.0),
         jnp.cos(theta / 2.0), 0],
        [-1.0j * jnp.sin(theta / 2.0), 0, 0,
         jnp.cos(theta / 2.0)],
    ])


#
# 3 Qubit Gates
#


@implements(getattr(qop, 'fredkin', None))
def fredkin(n=None, control=0, targets=[1, 2]):
    if [control, targets[0], targets[1]] != [0, 1, 2] and n is None:
        n = 3

    if n is not None:
        return gate_expand_3toN(fredkin(), n, [control, targets[0]], targets[1])
    return np.array([
        [1, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1],
    ])


@implements(getattr(qop, 'toffoli', None))
def toffoli(n=None, controls=[0, 1], target=2):
    if [controls[0], controls[1], target] != [0, 1, 2] and n is None:
        n = 3

    if n is not None:
        return gate_expand_3toN(toffoli(), n, controls, target)
    return np.array([
        [1, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0, 1, 0],
    ])


#
# Miscellaneous Gates
#


@implements(getattr(qop, 'rotation', None))
def rotation(op, phi, n=None, target=0):
    if n is not None:
        return gate_expand_1ton(rotation(op, phi), n, target)
    return jaxLA.expm(-1j * op * phi / 2)


@implements(getattr(qop, 'controlled_gate', None))
def controlled_gate(u, n=2, control=0, target=1, control_value=1):
    if [n, control, target] == [2, 0, 1]:
        return tensor(fock_dm(2, control_value), u) + tensor(
            fock_dm(2, 1 - control_value), identity(2))
    u2 = controlled_gate(u, control_value=control_value)
    return gate_expand_2ton(u2, n=n, control=control, target=target)


@implements(getattr(qop, 'globalphase', None))
def globalphase(theta, n=1):
    data = jnp.exp(1.0j * theta) * np.eye(2**n, 2**n, dtype=complex)
    return data


#
# Operation on Gates
#


@implements(getattr(qop, '_hamming_distance', None))
def _hamming_distance(x, bits=32):
    tot = 0
    while x:
        tot += 1
        x &= x - 1
    return tot


@implements(getattr(qop, 'hadamard_transform', None))
def hadamard_transform(n=1):
    data = [[1, 1], [1, -1]]
    H = np.array(data) / np.sqrt(2)
    return tensor(*([H] * n))


#
# Gate Expand
#


@implements(getattr(qop, 'gate_expand_1ton', None))
def gate_expand_1ton(u, n, target):
    if n < 1:
        raise ValueError("integer n must be larger or equal to 1")

    if target >= n:
        raise ValueError("target must be integer < integer N")
    return tensor(*([identity(2)] * (target) + [u] + [identity(2)] *
                    (n - target - 1)))


@implements(getattr(qop, 'gate_expand_2ton', None))
def gate_expand_2ton(u, n, control=None, target=None, targets=None):
    if targets is not None:
        control, target = targets

    if control is None or target is None:
        raise ValueError("Specify value of control and target")

    if n < 2:
        raise ValueError("integer n must be larger or equal to 2")

    if control >= n or target >= n:
        raise ValueError("control and not target must be integer < integer n")

    if control == target:
        raise ValueError("target and not control cannot be equal")

    p = list(range(n))

    if target == 0 and control == 1:
        p[control], p[target] = p[target], p[control]

    elif target == 0:
        p[1], p[target] = p[target], p[1]
        p[1], p[control] = p[control], p[1]

    else:
        p[1], p[target] = p[target], p[1]
        p[0], p[control] = p[control], p[0]

    return permute(tensor(*([u] + [identity(2)] * (n - 2))), [2] * n, p)


@implements(getattr(qop, 'gate_expand_3toN', None))
def gate_expand_3toN(u, n, controls=[0, 1], target=2):
    if n < 3:
        raise ValueError("integer n must be larger or equal to 3")

    if controls[0] >= n or controls[1] >= n or target >= n:
        raise ValueError("control and not target is None."
                         " Must be integer < integer N")

    if (controls[0] == target or controls[1] == target or
            controls[0] == controls[1]):

        raise ValueError("controls[0], controls[1], and target"
                         " cannot be equal")

    p = list(range(n))
    p1 = list(range(n))
    p2 = list(range(n))

    if controls[0] <= 2 and controls[1] <= 2 and target <= 2:
        p[controls[0]] = 0
        p[controls[1]] = 1
        p[target] = 2

    #
    # n > 3 cases
    #

    elif controls[0] == 0 and controls[1] == 1:
        p[2], p[target] = p[target], p[2]

    elif controls[0] == 0 and target == 2:
        p[1], p[controls[1]] = p[controls[1]], p[1]

    elif controls[1] == 1 and target == 2:
        p[0], p[controls[0]] = p[controls[0]], p[0]

    elif controls[0] == 1 and controls[1] == 0:
        p[controls[1]], p[controls[0]] = p[controls[0]], p[controls[1]]
        p2[2], p2[target] = p2[target], p2[2]
        p = [p2[p[k]] for k in range(n)]

    elif controls[0] == 2 and target == 0:
        p[target], p[controls[0]] = p[controls[0]], p[target]
        p1[1], p1[controls[1]] = p1[controls[1]], p1[1]
        p = [p1[p[k]] for k in range(n)]

    elif controls[1] == 2 and target == 1:
        p[target], p[controls[1]] = p[controls[1]], p[target]
        p1[0], p1[controls[0]] = p1[controls[0]], p1[0]
        p = [p1[p[k]] for k in range(n)]

    elif controls[0] == 1 and controls[1] == 2:
        #  controls[0] -> controls[1] -> target -> outside
        p[0], p[1] = p[1], p[0]
        p[0], p[2] = p[2], p[0]
        p[0], p[target] = p[target], p[0]

    elif controls[0] == 2 and target == 1:
        #  controls[0] -> target -> controls[1] -> outside
        p[0], p[2] = p[2], p[0]
        p[0], p[1] = p[1], p[0]
        p[0], p[controls[1]] = p[controls[1]], p[0]

    elif controls[1] == 0 and controls[0] == 2:
        #  controls[1] -> controls[0] -> target -> outside
        p[1], p[0] = p[0], p[1]
        p[1], p[2] = p[2], p[1]
        p[1], p[target] = p[target], p[1]

    elif controls[1] == 2 and target == 0:
        #  controls[1] -> target -> controls[0] -> outside
        p[1], p[2] = p[2], p[1]
        p[1], p[0] = p[0], p[1]
        p[1], p[controls[0]] = p[controls[0]], p[1]

    elif target == 1 and controls[1] == 0:
        #  target -> controls[1] -> controls[0] -> outside
        p[2], p[1] = p[1], p[2]
        p[2], p[0] = p[0], p[2]
        p[2], p[controls[0]] = p[controls[0]], p[2]

    elif target == 0 and controls[0] == 1:
        #  target -> controls[0] -> controls[1] -> outside
        p[2], p[0] = p[0], p[2]
        p[2], p[1] = p[1], p[2]
        p[2], p[controls[1]] = p[controls[1]], p[2]

    elif controls[0] == 0 and controls[1] == 2:
        #  controls[0] -> self, controls[1] -> target -> outside
        p[1], p[2] = p[2], p[1]
        p[1], p[target] = p[target], p[1]

    elif controls[1] == 1 and controls[0] == 2:
        #  controls[1] -> self, controls[0] -> target -> outside
        p[0], p[2] = p[2], p[0]
        p[0], p[target] = p[target], p[0]

    elif target == 2 and controls[0] == 1:
        #  target -> self, controls[0] -> controls[1] -> outside
        p[0], p[1] = p[1], p[0]
        p[0], p[controls[1]] = p[controls[1]], p[0]

    #
    # n > 4 cases
    #

    elif controls[0] == 1 and controls[1] > 2 and target > 2:
        #  controls[0] -> controls[1] -> outside, target -> outside
        p[0], p[1] = p[1], p[0]
        p[0], p[controls[1]] = p[controls[1]], p[0]
        p[2], p[target] = p[target], p[2]

    elif controls[0] == 2 and controls[1] > 2 and target > 2:
        #  controls[0] -> target -> outside, controls[1] -> outside
        p[0], p[2] = p[2], p[0]
        p[0], p[target] = p[target], p[0]
        p[1], p[controls[1]] = p[controls[1]], p[1]

    elif controls[1] == 2 and controls[0] > 2 and target > 2:
        #  controls[1] -> target -> outside, controls[0] -> outside
        p[1], p[2] = p[2], p[1]
        p[1], p[target] = p[target], p[1]
        p[0], p[controls[0]] = p[controls[0]], p[0]

    else:
        p[0], p[controls[0]] = p[controls[0]], p[0]
        p1[1], p1[controls[1]] = p1[controls[1]], p1[1]
        p2[2], p2[target] = p2[target], p2[2]
        p = [p[p1[p2[k]]] for k in range(n)]

    return permute(tensor(*([u] + [identity(2)] * (n - 3))), [2] * n, p)


@implements(getattr(qop, '_check_qubits_oper', None))
def _check_qubits_oper(oper, dims=None, targets=None):
    # if operator matches N
    if oper.dims[0] != oper.dims[1]:
        raise ValueError("The operator is not an "
                         "array with the same input and output dimensions.")
    # if operator dims matches the target dims
    if dims is not None and targets is not None:
        targ_dims = [dims[t] for t in targets]
        if oper.dims[0] != targ_dims:
            raise ValueError("The operator dims {} do not match "
                             "the target dims {}.".format(
                                 oper.dims[0], targ_dims))


@implements(getattr(qop, '_targets_to_list', None))
def _targets_to_list(targets, oper=None, n=None):
    # if targets is a list of integer
    if targets is None:
        targets = list(range(len(oper.dims[0])))
    if not isinstance(targets, Iterable):
        targets = [targets]
    if not all([isinstance(t, numbers.Integral) for t in targets]):
        raise TypeError("targets should be "
                        "an integer or a list of integer")
    # if targets has correct length
    if oper is not None:
        req_num = len(oper.dims[0])
        if len(targets) != req_num:
            raise ValueError("The given operator needs {} "
                             "target qutbis, "
                             "but {} given.".format(req_num, len(targets)))
    # if targets is smaller than N
    if n is not None:
        if not all([t < n for t in targets]):
            raise ValueError("Targets must be smaller than n={}.".format(n))
    return targets


@implements(getattr(qop, 'expand_operator', None))
def expand_operator(oper, n, targets, dims=None, cyclic_permutation=False):
    if dims is None:
        dims = [2] * n
    targets = _targets_to_list(targets, oper=oper, n=n)
    # _check_qubits_oper(oper, dims=dims, targets=targets)

    # Call expand_operator for all cyclic permutation of the targets.
    if cyclic_permutation:
        oper_list = []
        for i in range(n):
            new_targets = np.mod(np.array(targets) + i, n)
            oper_list.append(
                expand_operator(oper, n=n, targets=new_targets, dims=dims))
        return oper_list

    # Generate the correct order for qubits permutation,
    # eg. if n = 5, targets = [3,0], the order is [1,2,3,0,4].
    # If the operator is cnot,
    # this order means that the 3rd qubit controls the 0th qubit.
    new_order = [0] * n
    for i, t in enumerate(targets):
        new_order[t] = i
    # allocate the rest qutbits (not targets) to the empty
    # position in new_order
    rest_pos = [q for q in list(range(n)) if q not in targets]
    rest_qubits = list(range(len(targets), n))
    for i, ind in enumerate(rest_pos):
        new_order[ind] = rest_qubits[i]
    id_list = [identity(dims[i]) for i in rest_pos]
    return permute(tensor(*([oper] + id_list)), [2] * n, new_order)



import numpy as np
import jax.numpy as jnp

from supergrad.utils.fidelity import u_to_pauli


def single_qubit_gate_pauli_error(target_u,
                                  sim_u,
                                  depolarizing_rate=0.,
                                  target_count=None):
    """Diagnose the single qubit gate pauil error.

    Args:
        target_u: target unitary matrix
        sim_u: the unitary matrix from simulation
        depolarizing_rate (float): the depolarizing error rate (e.g. t_gate * r)

    Returns:
        jnp.ndarray: The cumulative weights of pauli errors depend on the number
            of gate error.
    """
    pe = u_to_pauli(jnp.conj(sim_u).T @ target_u)
    ind = np.unravel_index(np.arange(pe.size), pe.shape)
    pe = pe[ind]  # unravel index
    # check the number of gate error
    count = np.count_nonzero(np.array(ind), axis=0)
    # debug
    if target_count is not None:
        # debug port
        return pe[count == target_count], jnp.array(ind)[:,
                                                         count == target_count]
    crosstalk_error = []
    # weight-1 error
    crosstalk_error.append(jnp.sum(pe[count == 1]) / 6 + depolarizing_rate)
    # weight-2 error
    temp_ind = np.array(ind)[:, count == 2].T
    # pickup joint 2-gates error
    horizon = (np.sum(temp_ind[:, 2:], axis=1) == 0)
    horizon += (np.sum(temp_ind[:, (0, 1, 4, 5)], axis=1) == 0)
    horizon += (np.sum(temp_ind[:, :4], axis=1) == 0)
    vertical = (np.sum(temp_ind[:, (1, 3, 4, 5)], axis=1) == 0)
    vertical += (np.sum(temp_ind[:, (0, 2, 4, 5)], axis=1) == 0)
    vertical += (np.sum(temp_ind[:, (0, 1, 3, 5)], axis=1) == 0)
    vertical += (np.sum(temp_ind[:, (0, 1, 2, 4)], axis=1) == 0)
    idx = np.argwhere(horizon + vertical)
    error_list = pe[count == 2][idx]
    crosstalk_error.append(jnp.sum(error_list) / 7)
    # weight-3 error
    temp_ind = np.array(ind)[:, count == 3].T
    # pickup L-Type simultaneous 3-gates error
    upper_l = (np.sum(temp_ind[:, 4:], axis=1) == 0)
    lower_l = (np.sum(temp_ind[:, :2], axis=1) == 0)
    # pickup I-Type simultaneous 3-gates error
    left_i = (np.sum(temp_ind[:, (1, 3, 5)], axis=1) == 0)
    right_i = (np.sum(temp_ind[:, (0, 2, 4)], axis=1) == 0)
    idx = np.argwhere(upper_l + lower_l + left_i + right_i)
    error_list = pe[count == 3][idx]
    crosstalk_error.append(jnp.sum(error_list) / 10)

    return jnp.array(crosstalk_error)


def two_qubit_gate_pauli_error(target_u, sim_u, part_lst, depolarizing_rate=0.):
    """Diagnose the single qubit gate pauil error.

    Args:
        target_u: target unitary matrix
        sim_u: the unitary matrix from simulation
        part_list: list of tow qubit gate pair (renormalization part)
        depolarizing_rate (float): the depolarizing error rate (e.g. t_gate * r)

    Returns:
        jnp.ndarray: The cumulative weights of pauli errors depend on the number
            of gate error.
    """
    pe = u_to_pauli(jnp.conj(sim_u).T @ target_u)
    ind = np.unravel_index(np.arange(pe.size), pe.shape)
    pe = pe[ind]  # unravel index
    ind_array = np.array(ind)

    # Group the pauli error in two qubit pairs
    group_ind = []
    for part in part_lst:
        count = np.count_nonzero(ind_array[part, :], axis=0)
        new_ind = np.zeros_like(count)
        new_ind[np.nonzero(count)] = 1
        group_ind.append(new_ind)
    group_ind = np.array(group_ind)

    # check the number of gate error
    count = np.count_nonzero(group_ind, axis=0)
    crosstalk_error = []
    # weight-1 error
    crosstalk_error.append(jnp.sum(pe[count == 1]) / 3 + depolarizing_rate)
    # weight-2 error
    temp_ind = np.array(group_ind)[:, count == 2].T
    # pickup joint 2-gates error
    upper = (temp_ind[:, 2] == 0)
    lower = (temp_ind[:, 0] == 0)
    idx = np.argwhere(upper + lower)
    error_list = pe[count == 2][idx]
    crosstalk_error.append(jnp.sum(error_list) / 2)
    # weight-3 error
    crosstalk_error.append(jnp.sum(pe[count == 3]))

    return jnp.array(crosstalk_error)



import itertools
import numpy as np
import jax.numpy as jnp
import haiku as hk
import matplotlib.pyplot as plt

from supergrad.quantum_system.artificial_atom import Fluxonium

from qutip import Qobj
from qutip.qobj import ptrace
from qutip.entropy import entropy_vn


def state_entropy_vn(psi, sel):

    num_qubit = int(np.log2(psi.size))
    psi_qobj = Qobj(np.array(psi), dims=[[2] * num_qubit, [1] * num_qubit])
    dm = psi_qobj * psi_qobj.dag()
    partial_dm = ptrace(dm, sel)

    return entropy_vn(partial_dm, 2)


def inverse_participation_ratio(psi):
    """The inverse participation ratio discussed in
    (Transmon platform for quantum computing challenged by chaotic fluctuations) """

    return jnp.sum(jnp.abs(psi)**4)


def walsh_transform(energy_nd):
    """The Walsh-transform analysis.

    Return:
        coefficients, bitstrings
    """

    # truncation bosonic occupations
    trunc_slice = tuple(slice(0, 2) for _ in range(energy_nd.ndim))
    trunc_energy_nd = energy_nd[trunc_slice]
    # generate bitstring label
    bitstrings = [
        bit for bit in itertools.product((0, 1), repeat=energy_nd.ndim)
    ]
    # calculate the coefficients of the tau-Hamiltonian
    coeff_tau = []
    for bitstring in bitstrings:
        temp_sum = 0
        # Walsh-Hadamard transform
        for transform_b in bitstrings:
            temp_sum += (-1)**(np.array(bitstring) @ np.array(transform_b)
                               ) * trunc_energy_nd[transform_b]
        coeff_tau.append(temp_sum / 2**energy_nd.ndim)
    coeff_tau = np.abs(np.array(coeff_tau))
    # sort the coefficient
    sort_idx = np.argsort(-coeff_tau)  # in descending order
    return coeff_tau[sort_idx], np.array(bitstrings)[sort_idx]


def plot_walsh_transform(coeff, bitstring):
    """Plot the Walsh-transform coefficients."""

    # coeff, bitstring = walsh_transform(energy_nd)
    # change unit of coefficient
    coeff = coeff.copy() * 1e6 / 2 / np.pi
    # Sort weight of bitstring
    weight_2_coeff_lst = []
    weight_2_idx_lst = []
    weight_3_coeff_lst = []
    weight_3_idx_lst = []
    weight_other_coeff_lst = []
    weight_other_idx_lst = []
    for i in range(len(coeff)):
        if int(bitstring[i].sum()) == 2:
            weight_2_coeff_lst.append(coeff[i])
            weight_2_idx_lst.append(i)
        elif int(bitstring[i].sum()) == 3:
            weight_3_coeff_lst.append(coeff[i])
            weight_3_idx_lst.append(i)
        elif int(bitstring[i].sum()) > 3:
            weight_other_coeff_lst.append(coeff[i])
            weight_other_idx_lst.append(i)

    plt.scatter(weight_2_idx_lst,
                weight_2_coeff_lst,
                label=r'weight $w(\vec{b}) = 2$', s=60)
    plt.scatter(weight_3_idx_lst,
                weight_3_coeff_lst,
                label=r'weight $w(\vec{b}) = 3$', s=60)
    plt.scatter(weight_other_idx_lst,
                weight_other_coeff_lst,
                label=r'weight $w(\vec{b}) > 3$', s=60)
    plt.yscale('log')
    plt.xlabel(r'Bitstring $\vec{b}$')
    plt.xticks([])
    plt.yticks()
    plt.ylabel(r'$|c_\vec{b}|$ [kHz]')
    plt.legend()


def plot_subsystem_eigenenergies(params):
    """
    Plot eigenenergies for individual subsystems.
    TODO: implement for transmon, etc
    """

    @hk.without_apply_rng
    @hk.transform
    def fluxonium_eigenenergies():
        fluxonium = Fluxonium(phiext=0.5 * 2 * jnp.pi,
                              phi_max=5 * np.pi,
                              truncated_dim=10,
                              name='fluxonium')

        return fluxonium.eigenenergies()

    for subsys in params:
        rename_dict = {}
        rename_dict['fluxonium'] = params[subsys]
        ee = fluxonium_eigenenergies.apply(rename_dict)
        ee = ee - ee[0]
        plt.scatter(np.arange(3), ee[1:4], label=subsys)
        plt.legend()
    plt.show()


def print_pulse_frequencies(params, dict_key='omega_d', divide_2pi=True):

    for k in params:
        if dict_key in params[k]:
            if divide_2pi:
                print(k, float(params[k]['omega_d']) / 2 / np.pi)
            else:
                print(k, float(params[k]['omega_d']))


if __name__ == '__main__':
    params = {
        'fluxonium0': {
            'ec': jnp.array(1.),
            'ej': jnp.array(4.),
            'el': jnp.array(1.)
        },
        'fluxonium1': {
            'ec': jnp.array(1.),
            'ej': jnp.array(4.),
            'el': jnp.array(0.9)
        },
        'fluxonium2': {
            'ec': jnp.array(1.),
            'ej': jnp.array(4.),
            'el': jnp.array(1.1)
        },
        'fluxonium3': {
            'ec': jnp.array(1.),
            'ej': jnp.array(4.),
            'el': jnp.array(1.05)
        },
        'fluxonium4': {
            'ec': jnp.array(1.),
            'ej': jnp.array(4.),
            'el': jnp.array(0.95)
        }
    }
    plot_subsystem_eigenenergies(params)
